"""Motion state store and getters."""
from dataclasses import dataclass
from typing import List, Optional, Union

from opentrons.types import MountType, Point, StagingSlotName
from opentrons.hardware_control.types import CriticalPoint
from opentrons.motion_planning.adjacent_slots_getters import (
    get_east_west_slots,
    get_adjacent_slots,
)
from opentrons import motion_planning

from . import _move_types
from .. import errors
from ..types import (
    MotorAxis,
    WellLocation,
    LiquidHandlingWellLocation,
    CurrentWell,
    CurrentPipetteLocation,
    AddressableOffsetVector,
)
from .config import Config
from .labware import LabwareView
from .pipettes import PipetteView
from .addressable_areas import AddressableAreaView
from .geometry import GeometryView
from .modules import ModuleView
from .module_substates import HeaterShakerModuleId


@dataclass(frozen=True)
class PipetteLocationData:
    """Pipette data used to determine the current gantry position."""

    mount: MountType
    critical_point: Optional[CriticalPoint]


class MotionView:
    """Complete motion planning state and getter methods."""

    def __init__(
        self,
        config: Config,
        labware_view: LabwareView,
        pipette_view: PipetteView,
        addressable_area_view: AddressableAreaView,
        geometry_view: GeometryView,
        module_view: ModuleView,
    ) -> None:
        """Initialize a MotionState instance."""
        self._config = config
        self._labware = labware_view
        self._pipettes = pipette_view
        self._addressable_areas = addressable_area_view
        self._geometry = geometry_view
        self._modules = module_view

    def get_pipette_location(
        self,
        pipette_id: str,
        current_location: Optional[CurrentPipetteLocation] = None,
    ) -> PipetteLocationData:
        """Get the critical point of a pipette given the current location."""
        current_location = current_location or self._pipettes.get_current_location()
        pipette_data = self._pipettes.get(pipette_id)

        mount = pipette_data.mount
        critical_point = None

        # if the pipette was last used to move to a labware that requires
        # centering, set the critical point to the appropriate center
        if (
            isinstance(current_location, CurrentWell)
            and current_location.pipette_id == pipette_id
        ):
            if self._labware.get_should_center_column_on_target_well(
                current_location.labware_id
            ):
                critical_point = CriticalPoint.Y_CENTER
            elif self._labware.get_should_center_pipette_on_target_well(
                current_location.labware_id
            ):
                critical_point = CriticalPoint.XY_CENTER
        return PipetteLocationData(mount=mount, critical_point=critical_point)

    def get_movement_waypoints_to_well(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: Optional[Union[WellLocation, LiquidHandlingWellLocation]],
        origin: Point,
        origin_cp: Optional[CriticalPoint],
        max_travel_z: float,
        current_well: Optional[CurrentWell] = None,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        operation_volume: Optional[float] = None,
    ) -> List[motion_planning.Waypoint]:
        """Calculate waypoints to a destination that's specified as a well."""
        location = current_well or self._pipettes.get_current_location()

        destination_cp: Optional[CriticalPoint] = None
        if self._labware.get_should_center_column_on_target_well(labware_id):
            destination_cp = CriticalPoint.Y_CENTER
        elif self._labware.get_should_center_pipette_on_target_well(labware_id):
            destination_cp = CriticalPoint.XY_CENTER

        destination = self._geometry.get_well_position(
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            operation_volume=operation_volume,
            pipette_id=pipette_id,
        )

        move_type = _move_types.get_move_type_to_well(
            pipette_id, labware_id, well_name, location, force_direct
        )
        min_travel_z = self._geometry.get_min_travel_z(
            pipette_id, labware_id, location, minimum_z_height
        )

        destination_slot = self._geometry.get_ancestor_slot_name(labware_id)
        # TODO (spp, 11-29-2021): Should log some kind of warning that pipettes
        #  could crash onto the thermocycler if current well or addressable area is not known.
        extra_waypoints = self._geometry.get_extra_waypoints(
            location=location, to_slot=destination_slot
        )

        try:
            return motion_planning.get_waypoints(
                move_type=move_type,
                origin=origin,
                origin_cp=origin_cp,
                dest=destination,
                dest_cp=destination_cp,
                min_travel_z=min_travel_z,
                max_travel_z=max_travel_z,
                xy_waypoints=extra_waypoints,
            )
        except motion_planning.MotionPlanningError as error:
            raise errors.FailedToPlanMoveError(str(error))

    def get_movement_waypoints_to_addressable_area(
        self,
        addressable_area_name: str,
        offset: AddressableOffsetVector,
        origin: Point,
        origin_cp: Optional[CriticalPoint],
        max_travel_z: float,
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        stay_at_max_travel_z: bool = False,
        ignore_tip_configuration: Optional[bool] = True,
        max_travel_z_extra_margin: Optional[float] = None,
    ) -> List[motion_planning.Waypoint]:
        """Calculate waypoints to a destination that's specified as an addressable area."""
        location = self._pipettes.get_current_location()

        base_destination = (
            self._addressable_areas.get_addressable_area_move_to_location(
                addressable_area_name
            )
        )
        if stay_at_max_travel_z:
            base_destination_at_max_z = Point(
                base_destination.x,
                base_destination.y,
                # HACK(mm, 2023-12-18): We want to travel exactly at max_travel_z, but
                # motion_planning.get_waypoints() won't let us--the highest we can go is this margin
                # beneath max_travel_z. Investigate why motion_planning.get_waypoints() does not
                # let us travel at max_travel_z, and whether it's safe to make it do that.
                # Possibly related: https://github.com/Opentrons/opentrons/pull/6882#discussion_r514248062
                max_travel_z
                - motion_planning.waypoints.MINIMUM_Z_MARGIN
                - (max_travel_z_extra_margin or 0.0),
            )
            destination = base_destination_at_max_z + Point(
                offset.x, offset.y, offset.z
            )
        else:
            destination = base_destination + Point(offset.x, offset.y, offset.z)

        # TODO(jbl 11-28-2023) This may need to change for partial tip configurations on a 96
        if ignore_tip_configuration:
            destination_cp = CriticalPoint.INSTRUMENT_XY_CENTER
        else:
            destination_cp = CriticalPoint.XY_CENTER

        all_labware_highest_z = self._geometry.get_all_obstacle_highest_z()
        if minimum_z_height is None:
            minimum_z_height = float("-inf")
        min_travel_z = max(all_labware_highest_z, minimum_z_height)

        move_type = (
            motion_planning.MoveType.DIRECT
            if force_direct
            else motion_planning.MoveType.GENERAL_ARC
        )

        destination_slot = self._addressable_areas.get_addressable_area_base_slot(
            addressable_area_name
        )
        # TODO (spp, 11-29-2021): Should log some kind of warning that pipettes
        #  could crash onto the thermocycler if current well or addressable area is not known.
        extra_waypoints = self._geometry.get_extra_waypoints(
            location=location, to_slot=destination_slot
        )

        try:
            return motion_planning.get_waypoints(
                move_type=move_type,
                origin=origin,
                origin_cp=origin_cp,
                dest=destination,
                dest_cp=destination_cp,
                min_travel_z=min_travel_z,
                max_travel_z=max_travel_z,
                xy_waypoints=extra_waypoints,
            )
        except motion_planning.MotionPlanningError as error:
            raise errors.FailedToPlanMoveError(str(error))

    def get_movement_waypoints_to_coords(
        self,
        origin: Point,
        dest: Point,
        max_travel_z: float,
        direct: bool,
        additional_min_travel_z: Optional[float],
    ) -> List[motion_planning.Waypoint]:
        """Calculate waypoints to a destination that's specified as deck coordinates.

        Args:
            origin: The start point of the movement.
            dest: The end point of the movement.
            max_travel_z: How high, in deck coordinates, the pipette can go.
                This should be measured at the bottom of whatever tip is currently
                attached (if any).
            direct: If True, move directly. If False, move in an arc.
            additional_min_travel_z: The minimum height to clear, if moving in an arc.
                Ignored if `direct` is True. If lower than the default height,
                the default is used; this can only increase the height, not decrease it.
        """
        all_labware_highest_z = self._geometry.get_all_obstacle_highest_z()
        if additional_min_travel_z is None:
            additional_min_travel_z = float("-inf")
        min_travel_z = max(all_labware_highest_z, additional_min_travel_z)

        move_type = (
            motion_planning.MoveType.DIRECT
            if direct
            else motion_planning.MoveType.GENERAL_ARC
        )

        try:
            return motion_planning.get_waypoints(
                origin=origin,
                dest=dest,
                min_travel_z=min_travel_z,
                max_travel_z=max_travel_z,
                move_type=move_type,
                origin_cp=None,
                dest_cp=None,
            )
        except motion_planning.MotionPlanningError as error:
            raise errors.FailedToPlanMoveError(str(error))

    def check_pipette_blocking_hs_latch(
        self, hs_module_id: HeaterShakerModuleId
    ) -> bool:
        """Check if pipette would block h/s latch from opening if it is east, west or on module."""
        pipette_blocking = True
        current_location = self._pipettes.get_current_location()
        if current_location is not None:
            if isinstance(current_location, CurrentWell):
                ancestor = self._geometry.get_ancestor_slot_name(
                    current_location.labware_id
                )
                if isinstance(ancestor, StagingSlotName):
                    # Staging Area Slots cannot intersect with the h/s
                    return False
                pipette_deck_slot = ancestor.as_int()
            else:
                pipette_deck_slot = (
                    self._addressable_areas.get_addressable_area_base_slot(
                        current_location.addressable_area_name
                    ).as_int()
                )
            hs_deck_slot = self._modules.get_location(hs_module_id).slotName.as_int()
            conflicting_slots = get_east_west_slots(hs_deck_slot) + [hs_deck_slot]
            pipette_blocking = pipette_deck_slot in conflicting_slots
        return pipette_blocking

    def check_pipette_blocking_hs_shaker(
        self, hs_module_id: HeaterShakerModuleId
    ) -> bool:
        """Check if pipette would block h/s latch from starting shake if it is adjacent or on module."""
        pipette_blocking = True
        current_location = self._pipettes.get_current_location()
        if current_location is not None:
            if isinstance(current_location, CurrentWell):
                ancestor = self._geometry.get_ancestor_slot_name(
                    current_location.labware_id
                )
                if isinstance(ancestor, StagingSlotName):
                    # Staging Area Slots cannot intersect with the h/s
                    return False
                pipette_deck_slot = ancestor.as_int()
            else:
                pipette_deck_slot = (
                    self._addressable_areas.get_addressable_area_base_slot(
                        current_location.addressable_area_name
                    ).as_int()
                )
            hs_deck_slot = self._modules.get_location(hs_module_id).slotName.as_int()
            conflicting_slots = get_adjacent_slots(hs_deck_slot) + [hs_deck_slot]
            pipette_blocking = pipette_deck_slot in conflicting_slots
        return pipette_blocking

    def get_touch_tip_waypoints(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        center_point: Point,
        radius: float = 1.0,
    ) -> List[motion_planning.Waypoint]:
        """Get a list of touch points for a touch tip operation."""
        mount = self._pipettes.get_mount(pipette_id)
        labware_slot = self._geometry.get_ancestor_slot_name(labware_id)
        if isinstance(labware_slot, StagingSlotName):
            raise errors.LocationIsStagingSlotError(
                "Cannot perform Touch Tip on labware in Staging Area Slot."
            )
        next_to_module = self._modules.is_edge_move_unsafe(mount, labware_slot)
        edge_path_type = self._labware.get_edge_path_type(
            labware_id, well_name, mount, labware_slot, next_to_module
        )

        x_offset, y_offset = self._labware.get_well_radial_offsets(
            labware_id, well_name, radius
        )

        positions = _move_types.get_edge_point_list(
            center_point, x_offset, y_offset, edge_path_type
        )
        critical_point: Optional[CriticalPoint] = None

        if self._labware.get_should_center_column_on_target_well(labware_id):
            critical_point = CriticalPoint.Y_CENTER
        elif self._labware.get_should_center_pipette_on_target_well(labware_id):
            critical_point = CriticalPoint.XY_CENTER

        return [
            motion_planning.Waypoint(position=p, critical_point=critical_point)
            for p in positions
        ]

    def get_robot_mount_axes(self) -> List[MotorAxis]:
        """Get a list of axes belonging to all mounts on the robot."""
        mount_axes = [MotorAxis.LEFT_Z, MotorAxis.RIGHT_Z]
        if self._config.robot_type == "OT-3 Standard":
            mount_axes.append(MotorAxis.EXTENSION_Z)
        return mount_axes



"""Script for starting up emulation up with module emulators."""
import logging
import asyncio
from argparse import ArgumentParser
from typing import List

from opentrons.hardware_control.emulation.app import Application
from opentrons.hardware_control.emulation.scripts.run_module_emulator import (
    emulator_builder,
)
from opentrons.hardware_control.emulation.settings import Settings
from .run_module_emulator import run as run_module_by_name


async def run(settings: Settings, modules: List[str]) -> None:
    """Run the emulator app with connected module emulators.

    Args:
        settings: App settings.
        modules: The module emulators to start.

    Returns:
        None

    """
    loop = asyncio.get_event_loop()

    app_task = loop.create_task(Application(settings=settings).run())
    module_tasks = [
        loop.create_task(
            run_module_by_name(settings=settings, emulator_name=n, host="localhost")
        )
        for n in modules
    ]
    await asyncio.gather(app_task, *module_tasks)


def main() -> None:
    """Entry point."""
    a = ArgumentParser()
    a.add_argument(
        "--m",
        action="append",
        choices=emulator_builder.keys(),
        help="which module(s) to emulate.",
    )
    args = a.parse_args()

    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)
    asyncio.run(run(Settings(), args.m))


if __name__ == "__main__":
    main()



"""Customize the ProtocolEngine to monitor and control legacy (APIv2) protocols."""
from __future__ import annotations

import asyncio
from contextlib import ExitStack
from typing import Optional

from opentrons.legacy_commands.types import CommandMessage as LegacyCommand
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api.core.legacy.load_info import LoadInfo
from opentrons.protocol_engine import AbstractPlugin, actions as pe_actions
from opentrons.util.broker import ReadOnlyBroker

from .legacy_command_mapper import LegacyCommandMapper


class LegacyContextPlugin(AbstractPlugin):
    """A ProtocolEngine plugin to monitor and control an APIv2 protocol.

    In the legacy ProtocolContext, protocol execution is accomplished
    by direct communication with the HardwareControlAPI, as opposed to an
    intermediate layer like the ProtocolEngine. This plugin wraps up
    and hides this behavior, so the ProtocolEngine can monitor
    the run of a legacy protocol without affecting the execution of
    the protocol commands themselves.

    This plugin allows a ProtocolEngine to subscribe to what is being done with the
    legacy ProtocolContext, and insert matching commands into ProtocolEngine state for
    purely progress-tracking purposes.
    """

    def __init__(
        self,
        engine_loop: asyncio.AbstractEventLoop,
        broker: LegacyBroker,
        equipment_broker: ReadOnlyBroker[LoadInfo],
        legacy_command_mapper: Optional[LegacyCommandMapper] = None,
    ) -> None:
        """Initialize the plugin with its dependencies."""
        self._engine_loop = engine_loop

        self._broker = broker
        self._equipment_broker = equipment_broker
        self._legacy_command_mapper = legacy_command_mapper or LegacyCommandMapper()

        self._subscription_exit_stack: Optional[ExitStack] = None

    def setup(self) -> None:
        """Set up the plugin.

        Subscribe to the APIv2 context's message brokers to be informed
        of the APIv2 protocol's activity.
        """
        # Subscribe to activity on the APIv2 context,
        # and arrange to unsubscribe when this plugin is torn down.
        # Use an exit stack so if any part of this setup fails,
        # we clean up the parts that succeeded in reverse order.
        with ExitStack() as exit_stack:
            command_broker_unsubscribe = self._broker.subscribe(
                topic="command",
                handler=self._handle_legacy_command,
            )
            exit_stack.callback(command_broker_unsubscribe)

            exit_stack.enter_context(
                self._equipment_broker.subscribed(
                    callback=self._handle_equipment_loaded
                )
            )

            # All subscriptions succeeded.
            # Save the exit stack so our teardown method can use it later
            # to clean up these subscriptions.
            self._subscription_exit_stack = exit_stack.pop_all()

    # todo(mm, 2024-08-21): This no longer needs to be async.
    async def teardown(self) -> None:
        """Tear down the plugin, undoing the work done in `setup()`.

        Called by Protocol Engine.
        At this point, the APIv2 protocol script must have exited.
        """
        if self._subscription_exit_stack is not None:
            self._subscription_exit_stack.close()
            self._subscription_exit_stack = None

    def handle_action(self, action: pe_actions.Action) -> None:
        """React to a ProtocolEngine action."""
        # TODO(jbl 2022-07-06) handle_action stub should be completely removed
        pass

    def _handle_legacy_command(self, command: LegacyCommand) -> None:
        """Handle a command reported by the legacy APIv2 protocol.

        Used as a broker callback, so this will run in the APIv2 protocol's thread.
        """
        pe_actions = self._legacy_command_mapper.map_command(command=command)
        future = asyncio.run_coroutine_threadsafe(
            self._dispatch_action_list(pe_actions), self._engine_loop
        )
        future.result()

    def _handle_equipment_loaded(self, load_info: LoadInfo) -> None:
        """Handle an equipment load reported by the legacy APIv2 protocol.

        Used as a broker callback, so this will run in the APIv2 protocol's thread.
        """
        pe_actions = self._legacy_command_mapper.map_equipment_load(load_info=load_info)
        future = asyncio.run_coroutine_threadsafe(
            self._dispatch_action_list(pe_actions), self._engine_loop
        )
        future.result()

    async def _dispatch_action_list(self, actions: list[pe_actions.Action]) -> None:
        for action in actions:
            self.dispatch(action)



"""Base command data model and type definitions."""


from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Generic,
    Optional,
    TypeVar,
    List,
    Type,
    Union,
)

from pydantic import BaseModel, Field
from pydantic.generics import GenericModel

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.state.update_types import StateUpdate

from ..resources import ModelUtils
from ..errors import ErrorOccurrence
from ..notes import CommandNote, CommandNoteAdder

# Work around type-only circular dependencies.
if TYPE_CHECKING:
    from .. import execution
    from ..state.state import StateView


_ParamsT = TypeVar("_ParamsT", bound=BaseModel)
_ParamsT_contra = TypeVar("_ParamsT_contra", bound=BaseModel, contravariant=True)
_ResultT = TypeVar("_ResultT", bound=BaseModel)
_ResultT_co = TypeVar("_ResultT_co", bound=BaseModel, covariant=True)
_ErrorT = TypeVar("_ErrorT", bound=ErrorOccurrence)
_ErrorT_co = TypeVar("_ErrorT_co", bound=ErrorOccurrence, covariant=True)


class CommandStatus(str, Enum):
    """Command execution status."""

    QUEUED = "queued"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class CommandIntent(str, Enum):
    """Run intent for a given command.

    Props:
        PROTOCOL: the command is part of the protocol run itself.
        SETUP: the command is part of the setup phase of a run.
    """

    PROTOCOL = "protocol"
    SETUP = "setup"
    FIXIT = "fixit"


class BaseCommandCreate(
    GenericModel,
    # These type parameters need to be invariant because our fields are mutable.
    Generic[_ParamsT],
):
    """Base class for command creation requests.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    params: _ParamsT = Field(..., description="Command execution data payload")
    intent: Optional[CommandIntent] = Field(
        None,
        description=(
            "The reason the command was added. If not specified or `protocol`,"
            " the command will be treated as part of the protocol run itself,"
            " and added to the end of the existing command queue."
            "\n\n"
            "If `setup`, the command will be treated as part of run setup."
            " A setup command may only be enqueued if the run has not started."
            "\n\n"
            "Use setup commands for activities like pre-run calibration checks"
            " and module setup, like pre-heating."
        ),
    )
    key: Optional[str] = Field(
        None,
        description=(
            "A key value, unique in this run, that can be used to track"
            " the same logical command across multiple runs of the same protocol."
            " If a value is not provided, one will be generated."
        ),
    )


@dataclasses.dataclass(frozen=True)
class SuccessData(Generic[_ResultT_co]):
    """Data from the successful completion of a command."""

    public: _ResultT_co
    """Public result data. Exposed over HTTP and stored in databases."""

    state_update: StateUpdate = dataclasses.field(
        # todo(mm, 2024-08-22): Remove the default once all command implementations
        # use this, to make it harder to forget in new command implementations.
        default_factory=StateUpdate
    )
    """How the engine state should be updated to reflect this command success."""


@dataclasses.dataclass(frozen=True)
class DefinedErrorData(Generic[_ErrorT_co]):
    """Data from a command that failed with a defined error.

    This should only be used for "defined" errors, not any error.
    See `AbstractCommandImpl.execute()`.
    """

    public: _ErrorT_co
    """Public error data. Exposed over HTTP and stored in databases."""

    state_update: StateUpdate = dataclasses.field(
        # todo(mm, 2024-08-22): Remove the default once all command implementations
        # use this, to make it harder to forget in new command implementations.
        default_factory=StateUpdate
    )
    """How the engine state should be updated to reflect this command failure."""

    state_update_if_false_positive: StateUpdate = dataclasses.field(
        default_factory=StateUpdate
    )


class BaseCommand(
    GenericModel,
    # These type parameters need to be invariant because our fields are mutable.
    Generic[_ParamsT, _ResultT, _ErrorT],
):
    """Base command model.

    You shouldn't use this class directly; instead, use or define
    your own subclass per specific command type.
    """

    id: str = Field(
        ...,
        description="Unique identifier of this particular command instance",
    )
    createdAt: datetime = Field(..., description="Command creation timestamp")
    commandType: str = Field(
        ...,
        description=(
            "Specific command type that determines data requirements and "
            "execution behavior"
        ),
    )
    key: str = Field(
        ...,
        description=(
            "An identifier representing this command as a step in a protocol."
            " A command's `key` will be unique within a given run, but stable"
            " across all runs that perform the same exact procedure. Thus,"
            " `key` be used to compare/match commands across multiple runs"
            " of the same protocol."
        ),
    )
    status: CommandStatus = Field(..., description="Command execution status")
    params: _ParamsT = Field(..., description="Command execution data payload")
    result: Optional[_ResultT] = Field(
        None,
        description="Command execution result data, if succeeded",
    )
    error: Union[
        _ErrorT,
        # ErrorOccurrence here is a catch-all for undefined errors not captured by
        # _ErrorT, or defined errors that don't parse into _ErrorT because, for example,
        # they are from an older software version that was missing some fields.
        ErrorOccurrence,
        None,
    ] = Field(
        None,
        description="Reference to error occurrence, if execution failed",
    )
    startedAt: Optional[datetime] = Field(
        None,
        description="Command execution start timestamp, if started",
    )
    completedAt: Optional[datetime] = Field(
        None,
        description="Command execution completed timestamp, if completed",
    )
    intent: Optional[CommandIntent] = Field(
        None,
        description=(
            "The reason the command was added to the run."
            " If not specified or `protocol`, it is part of the protocol itself."
            " If `setup`, it was added as part of setup; for example,"
            " a command that is part of a calibration procedure."
        ),
    )
    notes: Optional[List[CommandNote]] = Field(
        None,
        description=(
            "Information not critical to the execution of the command derived from either"
            " the command's execution or the command's generation."
        ),
    )
    failedCommandId: Optional[str] = Field(
        None,
        description=(
            "FIXIT command use only. Reference of the failed command id we are trying to fix."
        ),
    )

    _ImplementationCls: Type[
        AbstractCommandImpl[
            _ParamsT,
            Union[
                SuccessData[
                    # Our _ImplementationCls must return public result data that can fit
                    # in our `result` field:
                    _ResultT,
                ],
                DefinedErrorData[
                    # Our _ImplementationCls must return public error data that can fit
                    # in our `error` field:
                    _ErrorT,
                ],
            ],
        ]
    ]


_ExecuteReturnT_co = TypeVar(
    "_ExecuteReturnT_co",
    bound=Union[
        SuccessData[BaseModel],
        DefinedErrorData[ErrorOccurrence],
    ],
    covariant=True,
)


class AbstractCommandImpl(
    ABC,
    Generic[_ParamsT_contra, _ExecuteReturnT_co],
):
    """Abstract command creation and execution implementation.

    A given command request should map to a specific command implementation,
    which defines how to execute the command and map data from execution into the
    result model.
    """

    def __init__(
        self,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        equipment: execution.EquipmentHandler,
        file_provider: execution.FileProvider,
        movement: execution.MovementHandler,
        gantry_mover: execution.GantryMover,
        labware_movement: execution.LabwareMovementHandler,
        pipetting: execution.PipettingHandler,
        tip_handler: execution.TipHandler,
        run_control: execution.RunControlHandler,
        rail_lights: execution.RailLightsHandler,
        model_utils: ModelUtils,
        status_bar: execution.StatusBarHandler,
        command_note_adder: CommandNoteAdder,
    ) -> None:
        """Initialize the command implementation with execution handlers."""
        pass

    @abstractmethod
    async def execute(self, params: _ParamsT_contra) -> _ExecuteReturnT_co:
        """Execute the command, mapping data from execution into a response model.

        This should either:

        - Return a `SuccessData`, if the command completed normally.
        - Return a `DefinedErrorData`, if the command failed with a "defined error."
          Defined errors are errors that are documented as part of the robot's public
          API.
        - Raise an exception, if the command failed with any other error
          (in other words, an undefined error).
        """
        ...



"""Protocol run control and management.

The main export of this module is the AbstractRunner class. See
protocol_runner.py for more details.
"""
from .protocol_runner import (
    AbstractRunner,
    RunResult,
    create_protocol_runner,
    JsonRunner,
    PythonAndLegacyRunner,
    LiveRunner,
    AnyRunner,
)
from .run_orchestrator import RunOrchestrator

__all__ = [
    "AbstractRunner",
    "RunResult",
    "create_protocol_runner",
    "JsonRunner",
    "PythonAndLegacyRunner",
    "LiveRunner",
    "AnyRunner",
    "RunOrchestrator",
]



"""Helpers for flagging unsafe movements around a Heater-Shaker Module."""

from typing import List, Optional

from opentrons.motion_planning.adjacent_slots_getters import (
    get_east_west_slots,
    get_north_south_slots,
)

from ..errors import (
    PipetteMovementRestrictedByHeaterShakerError,
    HeaterShakerLabwareLatchNotOpenError,
    HeaterShakerLabwareLatchStatusUnknown,
    WrongModuleTypeError,
)
from ..state.state import StateStore
from ..state.module_substates import HeaterShakerModuleSubState
from ..types import (
    HeaterShakerMovementRestrictors,
    HeaterShakerLatchStatus,
    LabwareLocation,
    ModuleLocation,
)
from ...hardware_control import HardwareControlAPI
from ...hardware_control.modules import HeaterShaker as HardwareHeaterShaker
from ...drivers.types import (
    HeaterShakerLabwareLatchStatus as HeaterShakerHardwareLatchStatus,
)


class HeaterShakerMovementFlagger:
    """A helper for flagging unsafe movements to a Heater-Shaker."""

    def __init__(
        self, state_store: StateStore, hardware_api: HardwareControlAPI
    ) -> None:
        self._state_store = state_store
        self._hardware_api = hardware_api

    async def raise_if_labware_latched_on_heater_shaker(
        self, labware_parent: LabwareLocation
    ) -> None:
        """Flag unsafe movements to a heater-shaker.

        If the given labware is on a heater-shaker, and that heater-shaker's labware
        latch is not open according to the engine's h/s state as well as the
        hardware API (for non-virtual modules), raises HeaterShakerLabwareLatchNotOpenError.
        If it is a virtual module, checks only for heater-shaker latch state in engine.

        Otherwise, no-ops.
        """
        if isinstance(labware_parent, ModuleLocation):
            module_id = labware_parent.moduleId
        else:
            return  # Labware not on a module.
        try:
            hs_substate = self._state_store.modules.get_heater_shaker_module_substate(
                module_id=module_id
            )
        except WrongModuleTypeError:
            return  # Labware on a module, but not a Heater-Shaker.

        if hs_substate.labware_latch_status == HeaterShakerLatchStatus.CLOSED:
            raise HeaterShakerLabwareLatchNotOpenError(
                "Heater-Shaker labware latch must be open when moving labware to/from it."
            )
        elif hs_substate.labware_latch_status == HeaterShakerLatchStatus.UNKNOWN:
            raise HeaterShakerLabwareLatchStatusUnknown(
                "Heater-Shaker labware latch must be opened before moving labware to/from it."
            )

        # There is a chance that the engine might not have the latest latch status;
        # do a hardware state check to be sure that the latch is truly open
        if not self._state_store.config.use_virtual_modules:
            await self._check_hardware_module_latch_status(hs_substate)

    async def _check_hardware_module_latch_status(
        self, hs_substate: HeaterShakerModuleSubState
    ) -> None:
        try:
            hs_latch_status = await self._get_hardware_heater_shaker_latch_status(
                module_id=hs_substate.module_id
            )
        except self._HardwareHeaterShakerMissingError as e:
            raise HeaterShakerLabwareLatchNotOpenError(
                "Heater-Shaker labware latch must be open when moving labware to/from it,"
                " but the latch status is unknown."
            ) from e

        if hs_latch_status != HeaterShakerHardwareLatchStatus.IDLE_OPEN:
            raise HeaterShakerLabwareLatchNotOpenError(
                f"Heater-Shaker labware latch must be open when moving labware to/from it,"
                f" but the latch is currently {hs_latch_status}"
            )

    async def _get_hardware_heater_shaker_latch_status(
        self,
        module_id: str,
    ) -> HeaterShakerHardwareLatchStatus:
        """Get latch status of the hardware H/S corresponding with the module ID.

        Returns:
            Latch status of the requested attached heater-shaker.

        Raises:
            _HardwareHeaterShakerMissingError: If we can't find that H/S in
                the hardware API, so we can't fetch its current latch status.
                It's unclear if this can happen in practice...
                maybe if the module disconnects between when it was loaded into
                Protocol Engine and when this function is called?
        """
        hs_serial = self._state_store.modules.get_serial_number(module_id=module_id)
        heater_shaker = await self._find_heater_shaker_by_serial(
            serial_number=hs_serial
        )
        if heater_shaker is None:
            raise self._HardwareHeaterShakerMissingError(
                f"No Heater-Shaker found with serial number {hs_serial}"
            )

        latch_status = heater_shaker.labware_latch_status
        # An attached H/S should always have latch status unless it's in error or
        # it was just connected and hasn't been polled for status yet.
        assert latch_status is not None, (
            "Did not receive a valid latch status from heater-shaker. "
            "Cannot verify safe labware movement."
        )
        return latch_status

    async def _find_heater_shaker_by_serial(
        self, serial_number: str
    ) -> Optional[HardwareHeaterShaker]:
        """Find the hardware H/S with the given serial number.

        Returns:
            The matching hardware H/S, or None if none was found.
        """
        for module in self._hardware_api.attached_modules:
            # Different module types have different keys under .device_info.
            # Heater-Shaker should always have .device_info["serial"].
            if (
                isinstance(module, HardwareHeaterShaker)
                and module.device_info["serial"] == serial_number
            ):
                return module
        return None

    def raise_if_movement_restricted(
        self,
        hs_movement_restrictors: List[HeaterShakerMovementRestrictors],
        destination_slot: int,
        is_multi_channel: bool,
        destination_is_tip_rack: bool,
    ) -> None:
        """Flag restricted movement around/to a Heater-Shaker."""
        for hs_movement_restrictor in hs_movement_restrictors:
            dest_east_west = destination_slot in get_east_west_slots(
                hs_movement_restrictor.deck_slot
            )
            dest_north_south = destination_slot in get_north_south_slots(
                hs_movement_restrictor.deck_slot
            )
            dest_heater_shaker = destination_slot == hs_movement_restrictor.deck_slot

            # If Heater-Shaker is running, can't move to or around it
            if (
                any([dest_east_west, dest_north_south, dest_heater_shaker])
                and hs_movement_restrictor.plate_shaking
            ):
                raise PipetteMovementRestrictedByHeaterShakerError(
                    "Cannot move pipette to Heater-Shaker or adjacent slot while module is shaking"
                )

            # If Heater-Shaker's latch is open, can't move to it or east and west of it
            elif hs_movement_restrictor.latch_status != HeaterShakerLatchStatus.CLOSED:
                if dest_heater_shaker:
                    raise PipetteMovementRestrictedByHeaterShakerError(
                        f"Cannot move pipette to Heater-Shaker while labware latch"
                        f" {'has not been closed' if hs_movement_restrictor.latch_status == HeaterShakerLatchStatus.UNKNOWN else 'is open'}."
                    )
                if (
                    dest_east_west
                    and self._state_store.config.robot_type == "OT-2 Standard"
                ):
                    raise PipetteMovementRestrictedByHeaterShakerError(
                        "Cannot move pipette to left or right of Heater-Shaker while labware latch "
                        f" {'has not been closed' if hs_movement_restrictor.latch_status == HeaterShakerLatchStatus.UNKNOWN else 'is open'}."
                    )

            elif (
                is_multi_channel
                and self._state_store.config.robot_type == "OT-2 Standard"
            ):
                # Can't go to east/west slot under any circumstances on OT-2
                # if pipette is multi-channel
                if dest_east_west:
                    raise PipetteMovementRestrictedByHeaterShakerError(
                        "Cannot move 8-Channel pipette to slot adjacent to the left or right of Heater-Shaker"
                    )
                # Can only go north/south if the labware is a tip rack
                elif dest_north_south and not destination_is_tip_rack:
                    raise PipetteMovementRestrictedByHeaterShakerError(
                        "Cannot move 8-Channel pipette to non-tip-rack labware directly in front of or behind a Heater-Shaker"
                    )

    class _HardwareHeaterShakerMissingError(Exception):
        pass



"""Basic protocol engine create file data state and store."""
from dataclasses import dataclass
from typing import List

from ._abstract_store import HasState, HandlesActions
from ..actions import Action, SucceedCommandAction
from ..commands import (
    Command,
    absorbance_reader,
)


@dataclass
class FileState:
    """State of Engine created files."""

    file_ids: List[str]


class FileStore(HasState[FileState], HandlesActions):
    """File state container."""

    _state: FileState

    def __init__(self) -> None:
        """Initialize a File store and its state."""
        self._state = FileState(file_ids=[])

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        if isinstance(action, SucceedCommandAction):
            self._handle_command(action.command)

    def _handle_command(self, command: Command) -> None:
        if isinstance(command.result, absorbance_reader.ReadAbsorbanceResult):
            if command.result.fileIds is not None:
                self._state.file_ids.extend(command.result.fileIds)


class FileView(HasState[FileState]):
    """Read-only engine created file state view."""

    _state: FileState

    def __init__(self, state: FileState) -> None:
        """Initialize the view of file state.

        Arguments:
            state: File state dataclass used for tracking file creation status.
        """
        self._state = state

    def get_filecount(self) -> int:
        """Get the number of files currently created by the protocol."""
        return len(self._state.file_ids)

    def get_file_id_list(self) -> List[str]:
        """Get the list of files by file ID created by the protocol."""
        return self._state.file_ids



from typing import Dict, Tuple, List, Optional

from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.protocol_engine.state.labware import LabwareLoadParams


# Default versions of Opentrons standard labware definitions in Python Protocol API
# v2.14 and above. Labware not explicitly listed here default to 1.
#
# TODO(jbl 2023-08-01) this needs to be done more holistically, both to find the version and make sure that
#   it corresponds to the API level is was released with
_APILEVEL_2_14_OT_DEFAULT_VERSIONS: Dict[str, int] = {
    # v1 of many labware definitions have wrong `zDimension`s. (Jira RSS-202.)
    # For "opentrons_96_aluminumblock_generic_pcr_strip_200ul" and
    # "opentrons_24_aluminumblock_generic_2ml_screwcap", they're wrong enough to
    # easily cause collisions. (Jira RSS-197.)
    "opentrons_24_aluminumblock_generic_2ml_screwcap": 2,
    "opentrons_96_aluminumblock_generic_pcr_strip_200ul": 2,
    # The following labware definitions have had a version bump due to using new properties
    # introduced in an inplace schema v2 update
    "armadillo_96_wellplate_200ul_pcr_full_skirt": 2,
    "biorad_96_wellplate_200ul_pcr": 2,
    "biorad_384_wellplate_50ul": 2,
    "corning_12_wellplate_6.9ml_flat": 2,
    "corning_384_wellplate_112ul_flat": 2,
    "corning_48_wellplate_1.6ml_flat": 2,
    "corning_96_wellplate_360ul_flat": 2,
    "nest_1_reservoir_195ml": 2,
    "nest_96_wellplate_100ul_pcr_full_skirt": 2,
    "nest_96_wellplate_200ul_flat": 2,
    "nest_96_wellplate_2ml_deep": 2,
    "opentrons_96_wellplate_200ul_pcr_full_skirt": 2,
    "corning_6_wellplate_16.8ml_flat": 2,
    "corning_24_wellplate_3.4ml_flat": 2,
}


class AmbiguousLoadLabwareParamsError(RuntimeError):
    """Error raised when specific labware parameters cannot be found due to multiple matching labware definitions."""


def resolve(
    load_name: str,
    namespace: Optional[str],
    version: Optional[int],
    custom_load_labware_params: List[LabwareLoadParams],
) -> Tuple[str, int]:
    """Resolve the load labware parameters that best matches any custom labware, or default to opentrons standards

    Args:
        load_name: Load name of the labware.
        namespace: Optionally provided labware definition namespace
        version: Optionally provided labware definition version
        custom_load_labware_params: List of load labware parameters associated with custom labware that
                                    match given parameters

    Returns:
        A tuple of the resolved namespace and version
    """

    def matches_params(custom_params: LabwareLoadParams) -> bool:
        matches_load_name = custom_params.load_name == load_name
        matches_namespace = namespace is None or custom_params.namespace == namespace
        matches_version = version is None or custom_params.version == version
        return matches_load_name and matches_namespace and matches_version

    if namespace is not None and version is not None:
        return namespace, version

    filtered_custom_params = [
        params for params in custom_load_labware_params if matches_params(params)
    ]

    if not filtered_custom_params:
        # No custom labware matches the input, but some standard labware might.
        # Use the Opentrons defaults for anything not explicitly provided.
        #
        # If the provided namespace was OPENTRONS_NAMESPACE, there would have been no
        # custom labware matching that namespace, so we will always take this path in
        # that case.
        resolved_namespace = namespace if namespace is not None else OPENTRONS_NAMESPACE
        resolved_version = (
            version
            if version is not None
            else _get_default_version_for_standard_labware(load_name=load_name)
        )

    elif len(filtered_custom_params) > 1:
        # Multiple custom labware match the input.
        raise AmbiguousLoadLabwareParamsError(
            f"Multiple custom labware associated with load name {load_name}."
        )

    else:
        # Exactly one custom labware matches the input. Return it.
        resolved_namespace = filtered_custom_params[0].namespace
        resolved_version = filtered_custom_params[0].version

    return resolved_namespace, resolved_version


def _get_default_version_for_standard_labware(load_name: str) -> int:
    # We know the protocol is running at least apiLevel 2.14 by this point because
    # apiLevel 2.13 and below has its own separate code path for resolving labware.
    return _APILEVEL_2_14_OT_DEFAULT_VERSIONS.get(load_name, 1)



"""Aspirate command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from opentrons_shared_data.errors.exceptions import PipetteOverpressureError
from typing_extensions import Literal

from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    LiquidHandlingWellLocationMixin,
    BaseLiquidHandlingResult,
    DestinationPositionResult,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

from opentrons.hardware_control import HardwareControlAPI

from ..state.update_types import StateUpdate, CLEAR
from ..types import WellLocation, WellOrigin, CurrentWell, DeckPoint

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils
    from ..state.state import StateView
    from ..notes import CommandNoteAdder


AspirateCommandType = Literal["aspirate"]


class AspirateParams(
    PipetteIdMixin, AspirateVolumeMixin, FlowRateMixin, LiquidHandlingWellLocationMixin
):
    """Parameters required to aspirate from a specific well."""

    pass


class AspirateResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from execution of an Aspirate command."""

    pass


_ExecuteReturn = Union[
    SuccessData[AspirateResult],
    DefinedErrorData[OverpressureError],
]


class AspirateImplementation(AbstractCommandImpl[AspirateParams, _ExecuteReturn]):
    """Aspirate command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        movement: MovementHandler,
        command_note_adder: CommandNoteAdder,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._movement = movement
        self._command_note_adder = command_note_adder
        self._model_utils = model_utils

    async def execute(self, params: AspirateParams) -> _ExecuteReturn:
        """Move to and aspirate from the requested well.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        ready_to_aspirate = self._pipetting.get_is_ready_to_aspirate(
            pipette_id=pipette_id
        )

        current_well = None
        state_update = StateUpdate()

        if not ready_to_aspirate:
            await self._movement.move_to_well(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=WellLocation(origin=WellOrigin.TOP),
            )

            await self._pipetting.prepare_for_aspirate(pipette_id=pipette_id)

            # set our current deck location to the well now that we've made
            # an intermediate move for the "prepare for aspirate" step
            current_well = CurrentWell(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
            )

        position = await self._movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=params.wellLocation,
            current_well=current_well,
            operation_volume=-params.volume,
        )
        deck_point = DeckPoint.construct(x=position.x, y=position.y, z=position.z)
        state_update.set_pipette_location(
            pipette_id=pipette_id,
            new_labware_id=labware_id,
            new_well_name=well_name,
            new_deck_point=deck_point,
        )

        try:
            volume_aspirated = await self._pipetting.aspirate_in_place(
                pipette_id=pipette_id,
                volume=params.volume,
                flow_rate=params.flowRate,
                command_note_adder=self._command_note_adder,
            )
        except PipetteOverpressureError as e:
            state_update.set_liquid_operated(
                labware_id=labware_id,
                well_name=well_name,
                volume_added=CLEAR,
            )
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo={"retryLocation": (position.x, position.y, position.z)},
                ),
                state_update=state_update,
            )
        else:
            state_update.set_liquid_operated(
                labware_id=labware_id,
                well_name=well_name,
                volume_added=-volume_aspirated,
            )
            return SuccessData(
                public=AspirateResult(
                    volume=volume_aspirated,
                    position=deck_point,
                ),
                state_update=state_update,
            )


class Aspirate(BaseCommand[AspirateParams, AspirateResult, OverpressureError]):
    """Aspirate command model."""

    commandType: AspirateCommandType = "aspirate"
    params: AspirateParams
    result: Optional[AspirateResult]

    _ImplementationCls: Type[AspirateImplementation] = AspirateImplementation


class AspirateCreate(BaseCommandCreate[AspirateParams]):
    """Create aspirate command request model."""

    commandType: AspirateCommandType = "aspirate"
    params: AspirateParams

    _CommandCls: Type[Aspirate] = Aspirate



"""Legacy protocol simulation core.

One way of simulating a Protocol API v2 protocol (e.g. to detect equipment requirements)
is to simulate at the level of the hardware API. However, this is slow.

To speed things up, these classes can additionally take the place of a
Protocol API v2 context's implementation, to simulate certain things at the level of
the Protocol API, avoiding certain calls to the underlying hardware API.
"""
from .legacy_protocol_core import LegacyProtocolCoreSimulator
from .legacy_instrument_core import LegacyInstrumentCoreSimulator

__all__ = [
    "LegacyProtocolCoreSimulator",
    "LegacyInstrumentCoreSimulator",
]



"""Reload labware command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import EquipmentHandler


ReloadLabwareCommandType = Literal["reloadLabware"]


class ReloadLabwareParams(BaseModel):
    """Payload required to load a labware into a slot."""

    labwareId: str = Field(
        ..., description="The already-loaded labware instance to update."
    )


class ReloadLabwareResult(BaseModel):
    """Result data from the execution of a LoadLabware command."""

    labwareId: str = Field(
        ...,
        description="An ID to reference this labware in subsequent commands. Same as the one in the parameters.",
    )
    offsetId: Optional[str] = Field(
        # Default `None` instead of `...` so this field shows up as non-required in
        # OpenAPI. The server is allowed to omit it or make it null.
        None,
        description=(
            "An ID referencing the labware offset that will apply"
            " to the reloaded labware."
            " This offset will be in effect until the labware is moved"
            " with a `moveLabware` command."
            " Null or undefined means no offset applies,"
            " so the default of (0, 0, 0) will be used."
        ),
    )


class ReloadLabwareImplementation(
    AbstractCommandImpl[ReloadLabwareParams, SuccessData[ReloadLabwareResult]]
):
    """Reload labware command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    async def execute(
        self, params: ReloadLabwareParams
    ) -> SuccessData[ReloadLabwareResult]:
        """Reload the definition and calibration data for a specific labware."""
        reloaded_labware = await self._equipment.reload_labware(
            labware_id=params.labwareId,
        )

        state_update = StateUpdate()

        state_update.set_labware_location(
            labware_id=params.labwareId,
            new_location=reloaded_labware.location,
            new_offset_id=reloaded_labware.offsetId,
        )

        return SuccessData(
            public=ReloadLabwareResult(
                labwareId=params.labwareId,
                offsetId=reloaded_labware.offsetId,
            ),
            state_update=state_update,
        )


class ReloadLabware(
    BaseCommand[ReloadLabwareParams, ReloadLabwareResult, ErrorOccurrence]
):
    """Reload labware command resource model."""

    commandType: ReloadLabwareCommandType = "reloadLabware"
    params: ReloadLabwareParams
    result: Optional[ReloadLabwareResult]

    _ImplementationCls: Type[ReloadLabwareImplementation] = ReloadLabwareImplementation


class ReloadLabwareCreate(BaseCommandCreate[ReloadLabwareParams]):
    """Reload labware command creation request."""

    commandType: ReloadLabwareCommandType = "reloadLabware"
    params: ReloadLabwareParams

    _CommandCls: Type[ReloadLabware] = ReloadLabware



"""Blow-out command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from opentrons_shared_data.errors.exceptions import PipetteOverpressureError
from typing_extensions import Literal


from ..state.update_types import StateUpdate
from ..types import DeckPoint
from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    FlowRateMixin,
    WellLocationMixin,
    DestinationPositionResult,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..errors.error_occurrence import ErrorOccurrence

from opentrons.hardware_control import HardwareControlAPI


if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..state.state import StateView
    from ..resources import ModelUtils


BlowOutCommandType = Literal["blowout"]


class BlowOutParams(PipetteIdMixin, FlowRateMixin, WellLocationMixin):
    """Payload required to blow-out a specific well."""

    pass


class BlowOutResult(DestinationPositionResult):
    """Result data from the execution of a blow-out command."""

    pass


_ExecuteReturn = Union[
    SuccessData[BlowOutResult],
    DefinedErrorData[OverpressureError],
]


class BlowOutImplementation(AbstractCommandImpl[BlowOutParams, _ExecuteReturn]):
    """BlowOut command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._movement = movement
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._model_utils = model_utils

    async def execute(self, params: BlowOutParams) -> _ExecuteReturn:
        """Move to and blow-out the requested well."""
        state_update = StateUpdate()

        x, y, z = await self._movement.move_to_well(
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
        )
        deck_point = DeckPoint.construct(x=x, y=y, z=z)
        state_update.set_pipette_location(
            pipette_id=params.pipetteId,
            new_labware_id=params.labwareId,
            new_well_name=params.wellName,
            new_deck_point=deck_point,
        )
        try:
            await self._pipetting.blow_out_in_place(
                pipette_id=params.pipetteId, flow_rate=params.flowRate
            )
        except PipetteOverpressureError as e:
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo={
                        "retryLocation": (
                            x,
                            y,
                            z,
                        )
                    },
                ),
            )
        else:
            return SuccessData(
                public=BlowOutResult(position=deck_point),
                state_update=state_update,
            )


class BlowOut(BaseCommand[BlowOutParams, BlowOutResult, ErrorOccurrence]):
    """Blow-out command model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams
    result: Optional[BlowOutResult]

    _ImplementationCls: Type[BlowOutImplementation] = BlowOutImplementation


class BlowOutCreate(BaseCommandCreate[BlowOutParams]):
    """Create blow-out command request model."""

    commandType: BlowOutCommandType = "blowout"
    params: BlowOutParams

    _CommandCls: Type[BlowOut] = BlowOut



"""Protocol file reading interfaces."""

from .protocol_files_invalid_error import ProtocolFilesInvalidError
from .protocol_reader import ProtocolReader
from .input_file import AbstractInputFile
from .protocol_source import (
    ProtocolSource,
    ProtocolSourceFile,
    ProtocolFileRole,
    ProtocolType,
    JsonProtocolConfig,
    PythonProtocolConfig,
)
from .file_reader_writer import FileReaderWriter, BufferedFile
from .file_hasher import FileHasher
from .extract_labware_definitions import extract_labware_definitions

__all__ = [
    # main interfaces
    "ProtocolReader",
    "extract_labware_definitions",
    # input values
    "AbstractInputFile",
    # errors
    "ProtocolFilesInvalidError",
    # values and types
    "ProtocolSource",
    "ProtocolSourceFile",
    "ProtocolFileRole",
    "ProtocolType",
    "JsonProtocolConfig",
    "PythonProtocolConfig",
    "BufferedFile",
    # helpers
    "FileReaderWriter",
    "FileHasher",
]



"""Validation file for protocol engine commandsot."""
from __future__ import annotations
from typing import Optional

from opentrons.protocol_engine.errors import HardwareNotSupportedError
from opentrons.hardware_control.protocols.types import FlexRobotType

from opentrons.hardware_control import HardwareControlAPI, OT3HardwareControlAPI


def ensure_ot3_hardware(
    hardware_api: HardwareControlAPI,
    error_msg: Optional[str] = None,
) -> OT3HardwareControlAPI:
    """Validate that the HardwareControlAPI is of OT-3 instance."""
    if hardware_api.get_robot_type() == FlexRobotType:
        return hardware_api  # type: ignore

    raise HardwareNotSupportedError(
        error_msg or "This command is supported by OT-3 only."
    )



import asyncio
import logging
from typing import Optional

from opentrons.hardware_control.emulation.abstract_emulator import AbstractEmulator
from opentrons.hardware_control.emulation.connection_handler import ConnectionHandler

log = logging.getLogger(__name__)


async def run_emulator_client(
    host: str,
    port: int,
    emulator: AbstractEmulator,
    retries: int = 3,
    interval_seconds: float = 0.1,
) -> None:
    """Run an emulator as a client.

    Args:
        host: Host to connect to.
        port: Port to connect on.
        emulator: The emulator instance.
        retries: Number of retries on a failed connection attempt.
        interval_seconds: How long to wait between retries.

    Returns:
        None
    """
    log.info(f"Connecting to {emulator.__class__.__name__} at {host}:{port}")

    r: Optional[asyncio.StreamReader] = None
    w: Optional[asyncio.StreamWriter] = None
    for i in range(retries):
        try:
            r, w = await asyncio.open_connection(host, port)
            break
        except IOError:
            log.error(
                f"{emulator.__class__.__name__} failed to connect on "
                f"try {i + 1}. Retrying in {interval_seconds} seconds."
            )
            await asyncio.sleep(interval_seconds)

    if r is None or w is None:
        raise IOError(
            f"Failed to connect to {emulator.__class__.__name__} at "
            f"{host}:{port} after {retries} retries."
        )

    connection = ConnectionHandler(emulator)
    await connection(r, w)


async def run_emulator_server(host: str, port: int, emulator: AbstractEmulator) -> None:
    """Run an emulator as a server.

    Args:
        host: Host to listen on.
        port: Port to listen on.
        emulator: Emulator instance.

    Returns:
        None
    """
    log.info(f"Starting {emulator.__class__.__name__} at {host}:{port}")
    server = await asyncio.start_server(ConnectionHandler(emulator), host, port)
    await server.serve_forever()



import asyncio
import functools
from typing import (
    Set,
    TypeVar,
    Type,
    cast,
    Callable,
    Any,
    Coroutine,
    ParamSpec,
    Concatenate,
)
from .types import ExecutionState
from opentrons_shared_data.errors.exceptions import ExecutionCancelledError

TaskContents = TypeVar("TaskContents")


class ExecutionManager:
    """This class holds onto a flag that is up/set while the hardware
    hardware controller is running and down/cleared when the hardware
    controller is "paused".

    It also handles loop clean up through its cancel method.
    """

    def __init__(self) -> None:
        self._state: ExecutionState = ExecutionState.RUNNING
        self._condition = asyncio.Condition()
        # this Task would technically be parametrized with every different thing
        # that you could possible call register_cancellable_task on unfortunately
        # so it's not gonna get typechecked
        self._cancellable_tasks: Set["asyncio.Task[Any]"] = set()

    async def pause(self) -> None:
        async with self._condition:
            if self._state is ExecutionState.CANCELLED:
                raise ExecutionCancelledError
            else:
                self._state = ExecutionState.PAUSED

    async def resume(self) -> None:
        async with self._condition:
            if self._state is ExecutionState.CANCELLED:
                pass
            else:
                self._state = ExecutionState.RUNNING
                self._condition.notify_all()

    async def cancel(self) -> None:
        async with self._condition:
            self._state = ExecutionState.CANCELLED
            self._condition.notify_all()
            running_task = asyncio.current_task()
            for t in self._cancellable_tasks:
                if t is not running_task:
                    t.cancel()

    async def reset(self) -> None:
        async with self._condition:
            self._state = ExecutionState.RUNNING
            self._condition.notify_all()

    async def get_state(self) -> ExecutionState:
        async with self._condition:
            return self._state

    def register_cancellable_task(self, task: "asyncio.Task[TaskContents]") -> None:
        self._cancellable_tasks.add(task)
        task.add_done_callback(lambda t: self._cancellable_tasks.discard(t))

    async def wait_for_is_running(self) -> None:
        async with self._condition:
            if self._state == ExecutionState.PAUSED:
                await self._condition.wait()
                # type-ignore needed because this is a reentrant function and narrowing cannot
                # apply
                if self._state == ExecutionState.CANCELLED:  # type: ignore[comparison-overlap]
                    raise ExecutionCancelledError
            elif self._state == ExecutionState.CANCELLED:
                raise ExecutionCancelledError
            else:
                pass


SubclassInstance = TypeVar("SubclassInstance", bound="ExecutionManagerProvider")
DecoratedMethodParams = ParamSpec("DecoratedMethodParams")
DecoratedReturn = TypeVar("DecoratedReturn")


class ExecutionManagerProvider:
    """A mixin that provides an execution manager.

    By inheriting from this class, hardware controller APIs can get an
    execution manager and implemented methods that provide access to it.
    """

    def __init__(self, simulator: bool) -> None:
        self._em_simulate = simulator
        self._execution_manager = ExecutionManager()
        self._taskify_movement_execution: bool = False

    @property
    def taskify_movement_execution(self) -> bool:
        return self._taskify_movement_execution

    @taskify_movement_execution.setter
    def taskify_movement_execution(self, cancellable: bool) -> None:
        self._taskify_movement_execution = cancellable

    @property
    def execution_manager(self) -> ExecutionManager:
        return self._execution_manager

    @classmethod
    def wait_for_running(
        cls: Type["ExecutionManagerProvider"],
        decorated: Callable[
            Concatenate[SubclassInstance, DecoratedMethodParams],
            Coroutine[Any, Any, DecoratedReturn],
        ],
    ) -> Callable[
        Concatenate[SubclassInstance, DecoratedMethodParams],
        Coroutine[Any, Any, DecoratedReturn],
    ]:
        @functools.wraps(decorated)
        async def replace(
            inst: SubclassInstance,
            *args: DecoratedMethodParams.args,
            **kwargs: DecoratedMethodParams.kwargs,
        ) -> DecoratedReturn:
            if not inst._em_simulate:
                await inst.execution_manager.wait_for_is_running()
            if inst.taskify_movement_execution:
                # Running these functions inside cancellable tasks makes it easier and
                # faster to cancel protocol runs. In the higher, runner & engine layers,
                # a cancellation request triggers cancellation of the running move task
                # and hence, prevents any further communication with hardware.
                decorated_task: "asyncio.Task[DecoratedReturn]" = asyncio.create_task(
                    decorated(inst, *args, **kwargs)
                )
                inst.execution_manager.register_cancellable_task(decorated_task)
                return await decorated_task
            else:
                return await decorated(inst, *args, **kwargs)

        return cast(
            Callable[
                Concatenate[SubclassInstance, DecoratedMethodParams],
                Coroutine[Any, Any, DecoratedReturn],
            ],
            replace,
        )

    async def do_delay(self, duration_s: float) -> None:
        if not self._em_simulate:

            async def sleep_for_seconds(seconds: float) -> None:
                await asyncio.sleep(seconds)

            delay_task = asyncio.create_task(sleep_for_seconds(duration_s))
            self._execution_manager.register_cancellable_task(delay_task)
            await delay_task



"""
- Driver is responsible for providing an interface for the temp-deck
- Driver is the only system component that knows about the temp-deck's GCODES
  or how the temp-deck communications

- Driver is NOT responsible interpreting the temperatures or states in any way
  or knowing anything about what the device is being used for
"""

from __future__ import annotations

import asyncio
import logging
from typing import Dict, Optional
from enum import Enum

from opentrons.drivers import utils
from opentrons.drivers.types import Temperature
from opentrons.drivers.command_builder import CommandBuilder
from opentrons.drivers.asyncio.communication import SerialConnection
from opentrons.drivers.temp_deck.abstract import AbstractTempDeckDriver

log = logging.getLogger(__name__)

DEFAULT_TEMP_DECK_TIMEOUT = 1

DEFAULT_COMMAND_RETRIES = 3


class GCODE(str, Enum):
    GET_TEMP = "M105"
    SET_TEMP = "M104"
    DEVICE_INFO = "M115"
    DISENGAGE = "M18"
    PROGRAMMING_MODE = "dfu"


TEMP_DECK_BAUDRATE = 115200

TEMP_DECK_COMMAND_TERMINATOR = "\r\n\r\n"
TEMP_DECK_ACK = "ok\r\nok\r\n"


class TempDeckError(Exception):
    pass


class TempDeckDriver(AbstractTempDeckDriver):
    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> TempDeckDriver:
        """
        Create a temp deck driver.

        Args:
            port: port or url of temp deck
            loop: optional event loop

        Returns: driver
        """
        connection = await SerialConnection.create(
            port=port,
            baud_rate=TEMP_DECK_BAUDRATE,
            timeout=DEFAULT_TEMP_DECK_TIMEOUT,
            ack=TEMP_DECK_ACK,
            loop=loop,
            reset_buffer_before_write=False,
        )
        return cls(connection=connection)

    def __init__(self, connection: SerialConnection) -> None:
        """
        Construct a temp deck driver

        Args:
            connection: Connection to the temp deck
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to the temp deck."""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from temp deck"""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check connected state"""
        return await self._connection.is_open()

    async def deactivate(self) -> None:
        """
        Send disengage command to temp deck

        Returns: None
        """
        c = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.DISENGAGE
        )
        await self._send_command(command=c)

    async def set_temperature(self, celsius: float) -> None:
        """
        Send a set temperate command to temp deck

        Args:
            celsius: the target temperature

        Returns: None
        """
        c = (
            CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR)
            .add_gcode(gcode=GCODE.SET_TEMP)
            .add_float(
                prefix="S",
                value=celsius,
                precision=utils.TEMPDECK_GCODE_ROUNDING_PRECISION,
            )
        )
        await self._send_command(command=c)

    async def get_temperature(self) -> Temperature:
        """
        Send a get temperature command to the temp deck.

        Returns: Temperature object

        """
        c = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.GET_TEMP
        )
        response = await self._send_command(command=c)
        return utils.parse_temperature_response(
            temperature_string=response,
            rounding_val=utils.TEMPDECK_GCODE_ROUNDING_PRECISION,
        )

    async def get_device_info(self) -> Dict[str, str]:
        """
        Queries Temp-Deck for its build version, model, and serial number

        returns: dict
            Where keys are the strings 'version', 'model', and 'serial',
            and each value is a string identifier

            {
                'serial': '1aa11bb22',
                'model': '1aa11bb22',
                'version': '1aa11bb22'
            }

        Example input from Temp-Deck's serial response:
            "serial:aa11bb22 model:aa11bb22 version:aa11bb22"
        """
        c = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.DEVICE_INFO
        )
        response = await self._send_command(command=c)
        return utils.parse_device_information(device_info_string=response)

    async def enter_programming_mode(self) -> None:
        """
        Send command to enter programming mode.

        Returns: None
        """
        c = CommandBuilder(terminator=TEMP_DECK_COMMAND_TERMINATOR).add_gcode(
            gcode=GCODE.PROGRAMMING_MODE
        )
        await self._send_command(command=c)

    async def _send_command(self, command: CommandBuilder) -> str:
        """
        Send the command

        Args:
            command: command to send

        Returns:
            command response
        """
        response = await self._connection.send_command(
            command=command, retries=DEFAULT_COMMAND_RETRIES
        )
        return response



from typing import Dict, Optional
from typing_extensions import Protocol

from ..types import SubSystem, SubSystemState


class HardwareManager(Protocol):
    """Protocol specifying access to configuration plane elements of hardware."""

    def get_fw_version(self) -> str:
        """
        Return the firmware version of the connected hardware.
        """
        ...

    @property
    def fw_version(self) -> str:
        """
        Return the firmware version of the connected hardware.

        The version is a string retrieved directly from the attached hardware
        (or possibly simulator).
        """
        ...

    @property
    def board_revision(self) -> str:
        """
        Return the revision of the central hardware.
        """
        ...

    @property
    def attached_subsystems(self) -> Dict[SubSystem, SubSystemState]:
        """
        Get a view of the hardware subsystems currently attached to the machine.

        These subsystems are the programmable parts that underly and provide things like motion
        control and instruments. Depending on the machine, different subsystems may be available.
        In general, callers should not use the presence or absence of a specific subsystem to decide
        whether or not the hardware is operating properly.
        """
        ...

    async def get_serial_number(self) -> Optional[str]:
        """Get the robot serial number, if provisioned. If not provisioned, will be None."""
        ...



"""Basic labware data state and store."""
from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    NamedTuple,
    cast,
    Union,
    overload,
)

from opentrons.protocol_engine.state import update_types
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.gripper.constants import LABWARE_GRIP_FORCE
from opentrons_shared_data.labware.labware_definition import (
    LabwareRole,
    InnerWellGeometry,
)
from opentrons_shared_data.pipette.types import LabwareUri

from opentrons.types import DeckSlotName, StagingSlotName, MountType
from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.protocols.models import LabwareDefinition, WellDefinition
from opentrons.calibration_storage.helpers import uri_from_details

from .. import errors
from ..resources import DeckFixedLabware, labware_validation, fixture_validation
from ..types import (
    DeckSlotLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    NonStackedLocation,
    Dimensions,
    LabwareOffset,
    LabwareOffsetVector,
    LabwareOffsetLocation,
    LabwareLocation,
    LoadedLabware,
    ModuleLocation,
    ModuleModel,
    OverlapOffset,
    LabwareMovementOffsetData,
    OnDeckLabwareLocation,
    OFF_DECK_LOCATION,
)
from ..actions import (
    Action,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    get_state_updates,
)
from ._abstract_store import HasState, HandlesActions
from ._move_types import EdgePathType


# URIs of labware whose definitions accidentally specify an engage height
# in units of half-millimeters instead of millimeters.
_MAGDECK_HALF_MM_LABWARE = {
    "opentrons/biorad_96_wellplate_200ul_pcr/1",
    "opentrons/nest_96_wellplate_100ul_pcr_full_skirt/1",
    "opentrons/usascientific_96_wellplate_2.4ml_deep/1",
}

_RIGHT_SIDE_SLOTS = {
    # OT-2:
    DeckSlotName.FIXED_TRASH,
    DeckSlotName.SLOT_9,
    DeckSlotName.SLOT_6,
    DeckSlotName.SLOT_3,
    # OT-3:
    DeckSlotName.SLOT_A3,
    DeckSlotName.SLOT_B3,
    DeckSlotName.SLOT_C3,
    DeckSlotName.SLOT_D3,
}


# The max height of the labware that can fit in a plate reader
_PLATE_READER_MAX_LABWARE_Z_MM = 16


class LabwareLoadParams(NamedTuple):
    """Parameters required to load a labware in Protocol Engine."""

    load_name: str
    namespace: str
    version: int


@dataclass
class LabwareState:
    """State of all loaded labware resources."""

    # Indexed by LoadedLabware.id.
    # If a LoadedLabware here has a non-None offsetId,
    # it must point to an existing element of labware_offsets_by_id.
    labware_by_id: Dict[str, LoadedLabware]

    # Indexed by LabwareOffset.id.
    # We rely on Python 3.7+ preservation of dict insertion order.
    labware_offsets_by_id: Dict[str, LabwareOffset]

    definitions_by_uri: Dict[str, LabwareDefinition]
    deck_definition: DeckDefinitionV5


class LabwareStore(HasState[LabwareState], HandlesActions):
    """Labware state container."""

    _state: LabwareState

    def __init__(
        self,
        deck_definition: DeckDefinitionV5,
        deck_fixed_labware: Sequence[DeckFixedLabware],
    ) -> None:
        """Initialize a labware store and its state."""
        definitions_by_uri: Dict[str, LabwareDefinition] = {
            uri_from_details(
                load_name=fixed_labware.definition.parameters.loadName,
                namespace=fixed_labware.definition.namespace,
                version=fixed_labware.definition.version,
            ): fixed_labware.definition
            for fixed_labware in deck_fixed_labware
        }
        labware_by_id = {
            fixed_labware.labware_id: LoadedLabware.construct(
                id=fixed_labware.labware_id,
                location=fixed_labware.location,
                loadName=fixed_labware.definition.parameters.loadName,
                definitionUri=uri_from_details(
                    load_name=fixed_labware.definition.parameters.loadName,
                    namespace=fixed_labware.definition.namespace,
                    version=fixed_labware.definition.version,
                ),
                offsetId=None,
            )
            for fixed_labware in deck_fixed_labware
        }

        self._state = LabwareState(
            definitions_by_uri=definitions_by_uri,
            labware_offsets_by_id={},
            labware_by_id=labware_by_id,
            deck_definition=deck_definition,
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._add_loaded_labware(state_update)
            self._set_labware_location(state_update)

        if isinstance(action, AddLabwareOffsetAction):
            labware_offset = LabwareOffset.construct(
                id=action.labware_offset_id,
                createdAt=action.created_at,
                definitionUri=action.request.definitionUri,
                location=action.request.location,
                vector=action.request.vector,
            )
            self._add_labware_offset(labware_offset)

        elif isinstance(action, AddLabwareDefinitionAction):
            uri = uri_from_details(
                namespace=action.definition.namespace,
                load_name=action.definition.parameters.loadName,
                version=action.definition.version,
            )
            self._state.definitions_by_uri[uri] = action.definition

    def _add_labware_offset(self, labware_offset: LabwareOffset) -> None:
        """Add a new labware offset to state.

        `labware_offset.id` must not match any existing labware offset ID.
        `LoadLabwareCommand`s retain references to their corresponding labware offsets
        and expect them to be immutable.
        """
        assert labware_offset.id not in self._state.labware_offsets_by_id

        self._state.labware_offsets_by_id[labware_offset.id] = labware_offset

    def _add_loaded_labware(self, state_update: update_types.StateUpdate) -> None:
        loaded_labware_update = state_update.loaded_labware
        if loaded_labware_update != update_types.NO_CHANGE:
            # If the labware load refers to an offset, that offset must actually exist.
            if loaded_labware_update.offset_id is not None:
                assert (
                    loaded_labware_update.offset_id in self._state.labware_offsets_by_id
                )

            definition_uri = uri_from_details(
                namespace=loaded_labware_update.definition.namespace,
                load_name=loaded_labware_update.definition.parameters.loadName,
                version=loaded_labware_update.definition.version,
            )

            self._state.definitions_by_uri[
                definition_uri
            ] = loaded_labware_update.definition

            location = loaded_labware_update.new_location

            display_name = loaded_labware_update.display_name

            self._state.labware_by_id[
                loaded_labware_update.labware_id
            ] = LoadedLabware.construct(
                id=loaded_labware_update.labware_id,
                location=location,
                loadName=loaded_labware_update.definition.parameters.loadName,
                definitionUri=definition_uri,
                offsetId=loaded_labware_update.offset_id,
                displayName=display_name,
            )

    def _set_labware_location(self, state_update: update_types.StateUpdate) -> None:
        labware_location_update = state_update.labware_location
        if labware_location_update != update_types.NO_CHANGE:
            labware_id = labware_location_update.labware_id
            new_offset_id = labware_location_update.offset_id

            self._state.labware_by_id[labware_id].offsetId = new_offset_id

            if labware_location_update.new_location:
                new_location = labware_location_update.new_location

                if isinstance(new_location, AddressableAreaLocation) and (
                    fixture_validation.is_gripper_waste_chute(
                        new_location.addressableAreaName
                    )
                    or fixture_validation.is_trash(new_location.addressableAreaName)
                ):
                    # If a labware has been moved into a waste chute it's been chuted away and is now technically off deck
                    new_location = OFF_DECK_LOCATION

                self._state.labware_by_id[labware_id].location = new_location


class LabwareView(HasState[LabwareState]):
    """Read-only labware state view."""

    _state: LabwareState

    def __init__(self, state: LabwareState) -> None:
        """Initialize the computed view of labware state.

        Arguments:
            state: Labware state dataclass used for all calculations.
        """
        self._state = state

    def get(self, labware_id: str) -> LoadedLabware:
        """Get labware data by the labware's unique identifier."""
        try:
            return self._state.labware_by_id[labware_id]
        except KeyError as e:
            raise errors.LabwareNotLoadedError(
                f"Labware {labware_id} not found."
            ) from e

    def get_id_by_module(self, module_id: str) -> str:
        """Return the ID of the labware loaded on the given module."""
        for labware_id, labware in self.state.labware_by_id.items():
            if (
                isinstance(labware.location, ModuleLocation)
                and labware.location.moduleId == module_id
            ):
                return labware_id

        raise errors.exceptions.LabwareNotLoadedOnModuleError(
            "There is no labware loaded on this Module"
        )

    def get_id_by_labware(self, labware_id: str) -> str:
        """Return the ID of the labware loaded on the given labware."""
        for labware in self.state.labware_by_id.values():
            if (
                isinstance(labware.location, OnLabwareLocation)
                and labware.location.labwareId == labware_id
            ):
                return labware.id
        raise errors.exceptions.LabwareNotLoadedOnLabwareError(
            f"There is not labware loaded onto labware {labware_id}"
        )

    def raise_if_labware_has_labware_on_top(self, labware_id: str) -> None:
        """Raise if labware has another labware on top."""
        for labware in self._state.labware_by_id.values():
            if (
                isinstance(labware.location, OnLabwareLocation)
                and labware.location.labwareId == labware_id
            ):
                raise errors.LabwareIsInStackError(
                    f"Cannot move to labware {labware_id}, labware has other labware stacked on top."
                )

    def get_by_slot(
        self,
        slot_name: Union[DeckSlotName, StagingSlotName],
    ) -> Optional[LoadedLabware]:
        """Get the labware located in a given slot, if any."""
        loaded_labware = list(self._state.labware_by_id.values())

        for labware in loaded_labware:
            if (
                isinstance(labware.location, DeckSlotLocation)
                and labware.location.slotName.id == slot_name.id
            ) or (
                isinstance(labware.location, AddressableAreaLocation)
                and labware.location.addressableAreaName == slot_name.id
            ):
                return labware

        return None

    def get_by_addressable_area(
        self,
        addressable_area: str,
    ) -> Optional[LoadedLabware]:
        """Get the labware located in a given addressable area, if any."""
        loaded_labware = list(self._state.labware_by_id.values())

        for labware in loaded_labware:
            if (
                isinstance(labware.location, AddressableAreaLocation)
                and labware.location.addressableAreaName == addressable_area
            ):
                return labware

        return None

    def get_definition(self, labware_id: str) -> LabwareDefinition:
        """Get labware definition by the labware's unique identifier."""
        return self.get_definition_by_uri(
            LabwareUri(self.get(labware_id).definitionUri)
        )

    def get_user_specified_display_name(self, labware_id: str) -> Optional[str]:
        """Get the labware's user-specified display name, if set."""
        return self.get(labware_id).displayName

    def get_display_name(self, labware_id: str) -> str:
        """Get the labware's display name.

        If a user-specified display name exists, will return that, else will return
        display name from the definition.
        """
        return (
            self.get_user_specified_display_name(labware_id)
            or self.get_definition(labware_id).metadata.displayName
        )

    def get_deck_definition(self) -> DeckDefinitionV5:
        """Get the current deck definition."""
        return self._state.deck_definition

    def get_definition_by_uri(self, uri: LabwareUri) -> LabwareDefinition:
        """Get the labware definition matching loadName namespace and version."""
        try:
            return self._state.definitions_by_uri[uri]
        except KeyError as e:
            raise errors.LabwareDefinitionDoesNotExistError(
                f"Labware definition for matching {uri} not found."
            ) from e

    def get_loaded_labware_definitions(self) -> List[LabwareDefinition]:
        """Get all loaded labware definitions."""
        loaded_labware = self._state.labware_by_id.values()
        return [
            self.get_definition_by_uri(LabwareUri(labware.definitionUri))
            for labware in loaded_labware
        ]

    def find_custom_labware_load_params(self) -> List[LabwareLoadParams]:
        """Find all load labware parameters for custom labware definitions in state."""
        return [
            LabwareLoadParams(
                load_name=definition.parameters.loadName,
                namespace=definition.namespace,
                version=definition.version,
            )
            for definition in self._state.definitions_by_uri.values()
            if definition.namespace != OPENTRONS_NAMESPACE
        ]

    def get_location(self, labware_id: str) -> LabwareLocation:
        """Get labware location by the labware's unique identifier."""
        return self.get(labware_id).location

    def get_parent_location(self, labware_id: str) -> NonStackedLocation:
        """Get labware's non-labware parent location."""
        parent = self.get_location(labware_id)
        if isinstance(parent, OnLabwareLocation):
            return self.get_parent_location(parent.labwareId)
        return parent

    def get_labware_stack(
        self, labware_stack: List[LoadedLabware]
    ) -> List[LoadedLabware]:
        """Get the a stack of labware starting from a given labware or existing stack."""
        parent = self.get_location(labware_stack[-1].id)
        if isinstance(parent, OnLabwareLocation):
            labware_stack.append(self.get(parent.labwareId))
            return self.get_labware_stack(labware_stack)
        return labware_stack

    def get_all(self) -> List[LoadedLabware]:
        """Get a list of all labware entries in state."""
        return list(self._state.labware_by_id.values())

    def get_has_quirk(self, labware_id: str, quirk: str) -> bool:
        """Get if a labware has a certain quirk."""
        return quirk in self.get_quirks(labware_id=labware_id)

    def get_quirks(self, labware_id: str) -> List[str]:
        """Get a labware's quirks."""
        definition = self.get_definition(labware_id)
        return definition.parameters.quirks or []

    def get_should_center_column_on_target_well(self, labware_id: str) -> bool:
        """True if a pipette moving to this labware should center its active column on the target.

        This is true for labware that have wells spanning entire columns.
        """
        has_quirk = self.get_has_quirk(labware_id, "centerMultichannelOnWells")
        return has_quirk and (
            len(self.get_definition(labware_id).wells) > 1
            and len(self.get_definition(labware_id).wells) < 96
        )

    def get_labware_stacking_maximum(self, labware: LabwareDefinition) -> int:
        """Returns the maximum number of labware allowed in a stack for a given labware definition.

        If not defined within a labware, defaults to one.
        """
        stacking_quirks = {
            "stackingMaxFive": 5,
            "stackingMaxFour": 4,
            "stackingMaxThree": 3,
            "stackingMaxTwo": 2,
            "stackingMaxOne": 1,
            "stackingMaxZero": 0,
        }
        for quirk in stacking_quirks.keys():
            if (
                labware.parameters.quirks is not None
                and quirk in labware.parameters.quirks
            ):
                return stacking_quirks[quirk]
        return 1

    def get_should_center_pipette_on_target_well(self, labware_id: str) -> bool:
        """True if a pipette moving to a well of this labware should center its body on the target.

        This is true for 1-well reservoirs no matter the pipette, and for large plates.
        """
        has_quirk = self.get_has_quirk(labware_id, "centerMultichannelOnWells")
        return has_quirk and (
            len(self.get_definition(labware_id).wells) == 1
            or len(self.get_definition(labware_id).wells) >= 96
        )

    def get_well_definition(
        self,
        labware_id: str,
        well_name: Optional[str] = None,
    ) -> WellDefinition:
        """Get a well's definition by labware and well name.

        If `well_name` is omitted, the first well in the labware
        will be used.
        """
        definition = self.get_definition(labware_id)

        if well_name is None:
            well_name = definition.ordering[0][0]

        try:
            return definition.wells[well_name]
        except KeyError as e:
            raise errors.WellDoesNotExistError(
                f"{well_name} does not exist in {labware_id}."
            ) from e

    def get_well_geometry(
        self, labware_id: str, well_name: Optional[str] = None
    ) -> InnerWellGeometry:
        """Get a well's inner geometry by labware and well name."""
        labware_def = self.get_definition(labware_id)
        if labware_def.innerLabwareGeometry is None:
            raise errors.IncompleteLabwareDefinitionError(
                message=f"No innerLabwareGeometry found in labware definition for labware_id: {labware_id}."
            )
        well_def = self.get_well_definition(labware_id, well_name)
        well_id = well_def.geometryDefinitionId
        if well_id is None:
            raise errors.IncompleteWellDefinitionError(
                message=f"No geometryDefinitionId found in well definition for well: {well_name} in labware_id: {labware_id}"
            )
        else:
            well_geometry = labware_def.innerLabwareGeometry.get(well_id)
            if well_geometry is None:
                raise errors.IncompleteLabwareDefinitionError(
                    message=f"No innerLabwareGeometry found in labware definition for well_id: {well_id} in labware_id: {labware_id}"
                )
            return well_geometry

    def get_well_size(
        self, labware_id: str, well_name: str
    ) -> Tuple[float, float, float]:
        """Get a well's size in x, y, z dimensions based on its shape.

        Args:
            labware_id: Labware identifier.
            well_name: Name of well in labware.

        Returns:
            A tuple of dimensions in x, y, and z. If well is circular,
            the x and y dimensions will both be set to the diameter.
        """
        well_definition = self.get_well_definition(labware_id, well_name)

        if well_definition.diameter is not None:
            x_size = y_size = well_definition.diameter
        else:
            # If diameter is None we know these values will be floats
            x_size = cast(float, well_definition.xDimension)
            y_size = cast(float, well_definition.yDimension)

        return x_size, y_size, well_definition.depth

    def get_well_radial_offsets(
        self, labware_id: str, well_name: str, radius_percentage: float
    ) -> Tuple[float, float]:
        """Get x and y radius offsets modified by radius percentage."""
        x_size, y_size, z_size = self.get_well_size(labware_id, well_name)
        return (x_size / 2.0) * radius_percentage, (y_size / 2.0) * radius_percentage

    def get_edge_path_type(
        self,
        labware_id: str,
        well_name: str,
        mount: MountType,
        labware_slot: DeckSlotName,
        next_to_module: bool,
    ) -> EdgePathType:
        """Get the recommended edge path type based on well column, labware position and any neighboring modules."""
        labware_definition = self.get_definition(labware_id)
        left_column = labware_definition.ordering[0]
        right_column = labware_definition.ordering[-1]

        left_path_criteria = mount is MountType.RIGHT and well_name in left_column
        right_path_criteria = mount is MountType.LEFT and well_name in right_column
        labware_right_side = labware_slot in _RIGHT_SIDE_SLOTS

        if left_path_criteria and (next_to_module or labware_right_side):
            return EdgePathType.LEFT
        elif right_path_criteria and next_to_module:
            return EdgePathType.RIGHT
        else:
            return EdgePathType.DEFAULT

    def validate_liquid_allowed_in_labware(
        self, labware_id: str, wells: Mapping[str, Any]
    ) -> List[str]:
        """Check if wells associated to a labware_id has well by name and that labware is not tiprack."""
        labware_definition = self.get_definition(labware_id)
        labware_wells = labware_definition.wells
        contains_wells = all(well_name in labware_wells for well_name in iter(wells))
        if labware_definition.parameters.isTiprack:
            raise errors.LabwareIsTipRackError(
                f"Given labware: {labware_id} is a tiprack. Can not load liquid."
            )
        if LabwareRole.adapter in labware_definition.allowedRoles:
            raise errors.LabwareIsAdapterError(
                f"Given labware: {labware_id} is an adapter. Can not load liquid."
            )
        if not contains_wells:
            raise errors.WellDoesNotExistError(
                f"Some of the supplied wells do not match the labwareId: {labware_id}."
            )
        return list(wells)

    def get_tip_length(self, labware_id: str, overlap: float = 0) -> float:
        """Get the nominal tip length of a tip rack."""
        definition = self.get_definition(labware_id)
        if definition.parameters.tipLength is None:
            raise errors.LabwareIsNotTipRackError(
                f"Labware {labware_id} has no tip length defined."
            )

        return definition.parameters.tipLength - overlap

    def get_tip_drop_z_offset(
        self, labware_id: str, length_scale: float, additional_offset: float
    ) -> float:
        """Get the tip drop offset from the top of the well."""
        tip_length = self.get_tip_length(labware_id)
        return -tip_length * length_scale + additional_offset

    def get_definition_uri(self, labware_id: str) -> LabwareUri:
        """Get a labware's definition URI."""
        return LabwareUri(self.get(labware_id).definitionUri)

    def get_uri_from_definition(
        self,
        labware_definition: LabwareDefinition,
    ) -> LabwareUri:
        """Get a definition URI from a full labware definition."""
        return uri_from_details(
            load_name=labware_definition.parameters.loadName,
            namespace=labware_definition.namespace,
            version=labware_definition.version,
        )

    def is_tiprack(self, labware_id: str) -> bool:
        """Get whether labware is a tiprack."""
        definition = self.get_definition(labware_id)
        return definition.parameters.isTiprack

    def get_load_name(self, labware_id: str) -> str:
        """Get the labware's load name."""
        definition = self.get_definition(labware_id)
        return definition.parameters.loadName

    @overload
    def get_dimensions(self, *, labware_definition: LabwareDefinition) -> Dimensions:
        pass

    @overload
    def get_dimensions(self, *, labware_id: str) -> Dimensions:
        pass

    def get_dimensions(
        self,
        *,
        labware_definition: LabwareDefinition | None = None,
        labware_id: str | None = None,
    ) -> Dimensions:
        """Get the labware's dimensions."""
        if labware_definition is None:
            assert labware_id is not None  # From our @overloads.
            labware_definition = self.get_definition(labware_id)

        dims = labware_definition.dimensions

        return Dimensions(
            x=dims.xDimension,
            y=dims.yDimension,
            z=dims.zDimension,
        )

    def get_labware_overlap_offsets(
        self, definition: LabwareDefinition, below_labware_name: str
    ) -> OverlapOffset:
        """Get the labware's overlap with requested labware's load name."""
        if below_labware_name in definition.stackingOffsetWithLabware.keys():
            stacking_overlap = definition.stackingOffsetWithLabware.get(
                below_labware_name, OverlapOffset(x=0, y=0, z=0)
            )
        else:
            stacking_overlap = definition.stackingOffsetWithLabware.get(
                "default", OverlapOffset(x=0, y=0, z=0)
            )
        return OverlapOffset(
            x=stacking_overlap.x, y=stacking_overlap.y, z=stacking_overlap.z
        )

    def get_module_overlap_offsets(
        self, definition: LabwareDefinition, module_model: ModuleModel
    ) -> OverlapOffset:
        """Get the labware's overlap with requested module model."""
        stacking_overlap = definition.stackingOffsetWithModule.get(
            str(module_model.value)
        )
        if not stacking_overlap:
            if self._is_thermocycler_on_ot2(module_model):
                return OverlapOffset(x=0, y=0, z=10.7)
            else:
                return OverlapOffset(x=0, y=0, z=0)

        return OverlapOffset(
            x=stacking_overlap.x, y=stacking_overlap.y, z=stacking_overlap.z
        )

    def _is_thermocycler_on_ot2(self, module_model: ModuleModel) -> bool:
        """Whether the given module is a thermocycler with the current deck being an OT2 deck."""
        robot_model = self.get_deck_definition()["robot"]["model"]
        return (
            module_model
            in [ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2]
            and robot_model == "OT-2 Standard"
        )

    def get_default_magnet_height(self, module_id: str, offset: float) -> float:
        """Return a labware's default Magnetic Module engage height with added offset, if supplied.

        The returned value is measured in millimeters above the labware base plane.
        """
        labware_id = self.get_id_by_module(module_id)
        parameters = self.get_definition(labware_id).parameters
        default_engage_height = parameters.magneticModuleEngageHeight
        if (
            parameters.isMagneticModuleCompatible is False
            or default_engage_height is None
        ):
            raise errors.exceptions.NoMagnetEngageHeightError(
                "The labware loaded on this Magnetic Module"
                " does not have a default engage height."
            )

        if self._is_magnetic_module_uri_in_half_millimeter(labware_id):
            # TODO(mc, 2022-09-26): this value likely _also_ needs a few mm subtracted
            # https://opentrons.atlassian.net/browse/RSS-111
            calculated_height = default_engage_height / 2.0
        else:
            calculated_height = default_engage_height

        return calculated_height + offset

    def get_labware_offset_vector(self, labware_id: str) -> LabwareOffsetVector:
        """Get the labware's calibration offset."""
        offset_id = self.get(labware_id=labware_id).offsetId
        if offset_id is None:
            return LabwareOffsetVector(x=0, y=0, z=0)
        else:
            return self._state.labware_offsets_by_id[offset_id].vector

    def get_labware_offset(self, labware_offset_id: str) -> LabwareOffset:
        """Get a labware offset by the offset's unique ID.

        Raises:
            LabwareOffsetDoesNotExistError: If the given ID does not match any
                                            previously added offset.
        """
        try:
            return self._state.labware_offsets_by_id[labware_offset_id]
        except KeyError as e:
            raise errors.LabwareOffsetDoesNotExistError(
                f"Labware offset {labware_offset_id} not found."
            ) from e

    def get_labware_offsets(self) -> List[LabwareOffset]:
        """Get all labware offsets, in the order they were added."""
        return list(self._state.labware_offsets_by_id.values())

    # TODO: Make this slightly more ergonomic for the caller by
    # only returning the optional str ID, at the cost of baking redundant lookups
    # into the API?
    def find_applicable_labware_offset(
        self,
        definition_uri: str,
        location: LabwareOffsetLocation,
    ) -> Optional[LabwareOffset]:
        """Find a labware offset that applies to the given definition and location.

        Returns the *most recently* added matching offset,
        so later offsets can override earlier ones.
        Or, ``None`` if no offsets match at all.

        An offset "matches"
        if its ``definition_uri`` and ``location`` *exactly* match what's provided.
        This implies that if the location involves a module,
        it will *not* match a module that's compatible but not identical.
        """
        for candidate in reversed(list(self._state.labware_offsets_by_id.values())):
            if (
                candidate.definitionUri == definition_uri
                and candidate.location == location
            ):
                return candidate

        return None

    def get_fixed_trash_id(self) -> Optional[str]:
        """Get the identifier of labware loaded into the fixed trash location.

        Raises:
            LabwareNotLoadedError: a fixed trash was not loaded by the deck definition
                that is currently in use for the protocol run.
        """
        for labware in self._state.labware_by_id.values():
            if isinstance(
                labware.location, DeckSlotLocation
            ) and labware.location.slotName in {
                DeckSlotName.FIXED_TRASH,
                DeckSlotName.SLOT_A3,
            }:
                return labware.id
        return None

    def is_fixed_trash(self, labware_id: str) -> bool:
        """Check if labware is fixed trash."""
        return self.get_has_quirk(labware_id, "fixedTrash")

    def is_absorbance_reader_lid(self, labware_id: str) -> bool:
        """Check if labware is an absorbance reader lid."""
        return labware_validation.is_absorbance_reader_lid(
            self.get(labware_id).loadName
        )

    def raise_if_labware_inaccessible_by_pipette(self, labware_id: str) -> None:
        """Raise an error if the specified location cannot be reached via a pipette."""
        labware = self.get(labware_id)
        labware_location = labware.location
        if isinstance(labware_location, OnLabwareLocation):
            return self.raise_if_labware_inaccessible_by_pipette(
                labware_location.labwareId
            )
        elif isinstance(labware_location, AddressableAreaLocation):
            if fixture_validation.is_staging_slot(labware_location.addressableAreaName):
                raise errors.LocationNotAccessibleByPipetteError(
                    f"Cannot move pipette to {labware.loadName},"
                    f" labware is on staging slot {labware_location.addressableAreaName}"
                )
        elif labware_location == OFF_DECK_LOCATION:
            raise errors.LocationNotAccessibleByPipetteError(
                f"Cannot move pipette to {labware.loadName}, labware is off-deck."
            )

    def raise_if_labware_in_location(
        self,
        location: OnDeckLabwareLocation,
    ) -> None:
        """Raise an error if the specified location has labware in it."""
        for labware in self.get_all():
            if labware.location == location:
                raise errors.LocationIsOccupiedError(
                    f"Labware {labware.loadName} is already present at {location}."
                )

    def raise_if_labware_incompatible_with_plate_reader(
        self,
        labware_definition: LabwareDefinition,
    ) -> None:
        """Raise an error if the labware is not compatible with the plate reader."""
        load_name = labware_definition.parameters.loadName
        number_of_wells = len(labware_definition.wells)
        if number_of_wells != 96:
            raise errors.LabwareMovementNotAllowedError(
                f"Cannot move '{load_name}' into plate reader because the"
                f" labware contains {number_of_wells} wells where 96 wells is expected."
            )
        elif labware_definition.dimensions.zDimension > _PLATE_READER_MAX_LABWARE_Z_MM:
            raise errors.LabwareMovementNotAllowedError(
                f"Cannot move '{load_name}' into plate reader because the"
                f" maximum allowed labware height is {_PLATE_READER_MAX_LABWARE_Z_MM}mm."
            )

    def raise_if_labware_cannot_be_stacked(  # noqa: C901
        self, top_labware_definition: LabwareDefinition, bottom_labware_id: str
    ) -> None:
        """Raise if the specified labware definition cannot be placed on top of the bottom labware."""
        if labware_validation.validate_definition_is_adapter(top_labware_definition):
            raise errors.LabwareCannotBeStackedError(
                f"Labware {top_labware_definition.parameters.loadName} is defined as an adapter and cannot be placed"
                " on other labware."
            )
        below_labware = self.get(bottom_labware_id)
        if not labware_validation.validate_labware_can_be_stacked(
            top_labware_definition=top_labware_definition,
            below_labware_load_name=below_labware.loadName,
        ):
            raise errors.LabwareCannotBeStackedError(
                f"Labware {top_labware_definition.parameters.loadName} cannot be loaded onto labware {below_labware.loadName}"
            )
        elif isinstance(below_labware.location, ModuleLocation):
            below_definition = self.get_definition(labware_id=below_labware.id)
            if not labware_validation.validate_definition_is_adapter(
                below_definition
            ) and not labware_validation.validate_definition_is_lid(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded"
                    f" onto a labware on top of a module"
                )
        elif isinstance(below_labware.location, OnLabwareLocation):
            labware_stack = self.get_labware_stack([below_labware])
            stack_without_adapters = []
            for lw in labware_stack:
                if not labware_validation.validate_definition_is_adapter(
                    self.get_definition(lw.id)
                ):
                    stack_without_adapters.append(lw)
            if len(stack_without_adapters) >= self.get_labware_stacking_maximum(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded to stack of more than {self.get_labware_stacking_maximum(top_labware_definition)} labware."
                )

            further_below_definition = self.get_definition(
                labware_id=below_labware.location.labwareId
            )
            if labware_validation.validate_definition_is_adapter(
                further_below_definition
            ) and not labware_validation.validate_definition_is_lid(
                top_labware_definition
            ):
                raise errors.LabwareCannotBeStackedError(
                    f"Labware {top_labware_definition.parameters.loadName} cannot be loaded"
                    f" onto labware on top of adapter"
                )

    def _is_magnetic_module_uri_in_half_millimeter(self, labware_id: str) -> bool:
        """Check whether the labware uri needs to be calculated in half a millimeter."""
        uri = self.get_uri_from_definition(self.get_definition(labware_id))
        return uri in _MAGDECK_HALF_MM_LABWARE

    def get_deck_default_gripper_offsets(self) -> Optional[LabwareMovementOffsetData]:
        """Get the deck's default gripper offsets."""
        parsed_offsets = (
            self.get_deck_definition().get("gripperOffsets", {}).get("default")
        )
        return (
            LabwareMovementOffsetData(
                pickUpOffset=LabwareOffsetVector(
                    x=parsed_offsets["pickUpOffset"]["x"],
                    y=parsed_offsets["pickUpOffset"]["y"],
                    z=parsed_offsets["pickUpOffset"]["z"],
                ),
                dropOffset=LabwareOffsetVector(
                    x=parsed_offsets["dropOffset"]["x"],
                    y=parsed_offsets["dropOffset"]["y"],
                    z=parsed_offsets["dropOffset"]["z"],
                ),
            )
            if parsed_offsets
            else None
        )

    def get_absorbance_reader_lid_definition(self) -> LabwareDefinition:
        """Return the special labware definition for the plate reader lid.

        See todo comments in `create_protocol_engine().
        """
        # NOTE: This needs to stay in sync with create_protocol_engine().
        return self._state.definitions_by_uri[
            "opentrons/opentrons_flex_lid_absorbance_plate_reader_module/1"
        ]

    @overload
    def get_child_gripper_offsets(
        self,
        *,
        labware_definition: LabwareDefinition,
        slot_name: Optional[DeckSlotName],
    ) -> Optional[LabwareMovementOffsetData]:
        pass

    @overload
    def get_child_gripper_offsets(
        self, *, labware_id: str, slot_name: Optional[DeckSlotName]
    ) -> Optional[LabwareMovementOffsetData]:
        pass

    def get_child_gripper_offsets(
        self,
        *,
        labware_definition: Optional[LabwareDefinition] = None,
        labware_id: Optional[str] = None,
        slot_name: Optional[DeckSlotName],
    ) -> Optional[LabwareMovementOffsetData]:
        """Get the grip offsets that a labware says should be applied to children stacked atop it.

        Params:
            labware_id: The ID of a parent labware (atop which another labware, the child, will be stacked).
            slot_name: The ancestor slot that the parent labware is ultimately loaded into,
                       perhaps after going through a module in the middle.

        Returns:
            If `slot_name` is provided, returns the gripper offsets that the parent labware definition
            specifies just for that slot, or `None` if the labware definition doesn't have an
            exact match.

            If `slot_name` is `None`, returns the gripper offsets that the parent labware
            definition designates as "default," or `None` if it doesn't designate any as such.
        """
        if labware_id is not None:
            labware_definition = self.get_definition(labware_id)
        else:
            # Should be ensured by our @overloads.
            assert labware_definition is not None

        parsed_offsets = labware_definition.gripperOffsets
        offset_key = slot_name.id if slot_name else "default"

        if parsed_offsets is None or offset_key not in parsed_offsets:
            return None
        else:
            return LabwareMovementOffsetData(
                pickUpOffset=cast(
                    LabwareOffsetVector, parsed_offsets[offset_key].pickUpOffset
                ),
                dropOffset=cast(
                    LabwareOffsetVector, parsed_offsets[offset_key].dropOffset
                ),
            )

    def get_grip_force(self, labware_definition: LabwareDefinition) -> float:
        """Get the recommended grip force for gripping labware using gripper."""
        recommended_force = labware_definition.gripForce
        return (
            recommended_force if recommended_force is not None else LABWARE_GRIP_FORCE
        )

    def get_grip_height_from_labware_bottom(
        self, labware_definition: LabwareDefinition
    ) -> float:
        """Get the recommended grip height from labware bottom, if present."""
        recommended_height = labware_definition.gripHeightFromLabwareBottom
        return (
            recommended_height
            if recommended_height is not None
            else self.get_dimensions(labware_definition=labware_definition).z / 2
        )

    @staticmethod
    def _max_x_of_well(well_defn: WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.x + (well_defn.xDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.x + (well_defn.diameter or 0) / 2
        else:
            return well_defn.x

    @staticmethod
    def _min_x_of_well(well_defn: WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.x - (well_defn.xDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.x - (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _max_y_of_well(well_defn: WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.y + (well_defn.yDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.y + (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _min_y_of_well(well_defn: WellDefinition) -> float:
        if well_defn.shape == "rectangular":
            return well_defn.y - (well_defn.yDimension or 0) / 2
        elif well_defn.shape == "circular":
            return well_defn.y - (well_defn.diameter or 0) / 2
        else:
            return 0

    @staticmethod
    def _max_z_of_well(well_defn: WellDefinition) -> float:
        return well_defn.z + well_defn.depth

    def get_well_bbox(self, labware_definition: LabwareDefinition) -> Dimensions:
        """Get the bounding box implied by the wells.

        The bounding box of the labware that is implied by the wells is that required
        to contain the bounds of the wells - the y-span from the min-y bound of the min-y
        well to the max-y bound of the max-y well, x ditto, z from labware 0 to the max-z
        well top.

        This is used for the specific purpose of finding the reasonable uncertainty bounds of
        where and how a gripper will interact with a labware.
        """
        max_x: Optional[float] = None
        min_x: Optional[float] = None
        max_y: Optional[float] = None
        min_y: Optional[float] = None
        max_z: Optional[float] = None

        for well in labware_definition.wells.values():
            well_max_x = self._max_x_of_well(well)
            well_min_x = self._min_x_of_well(well)
            well_max_y = self._max_y_of_well(well)
            well_min_y = self._min_y_of_well(well)
            well_max_z = self._max_z_of_well(well)
            if (max_x is None) or (well_max_x > max_x):
                max_x = well_max_x
            if (max_y is None) or (well_max_y > max_y):
                max_y = well_max_y
            if (min_x is None) or (well_min_x < min_x):
                min_x = well_min_x
            if (min_y is None) or (well_min_y < min_y):
                min_y = well_min_y
            if (max_z is None) or (well_max_z > max_z):
                max_z = well_max_z
        if (
            max_x is None
            or max_y is None
            or min_x is None
            or min_y is None
            or max_z is None
        ):
            return Dimensions(0, 0, 0)
        return Dimensions(max_x - min_x, max_y - min_y, max_z)



"""Main ProtocolEngine factory."""
import asyncio
import contextlib
import typing

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import DoorState
from opentrons.protocol_engine.execution.error_recovery_hardware_state_synchronizer import (
    ErrorRecoveryHardwareStateSynchronizer,
)
from opentrons.protocol_engine.resources.labware_data_provider import (
    LabwareDataProvider,
)
from opentrons.util.async_helpers import async_context_manager_in_thread

from opentrons_shared_data.robot import load as load_robot

from .actions.action_dispatcher import ActionDispatcher
from .error_recovery_policy import ErrorRecoveryPolicy
from .execution.door_watcher import DoorWatcher
from .execution.hardware_stopper import HardwareStopper
from .plugins import PluginStarter
from .protocol_engine import ProtocolEngine
from .resources import DeckDataProvider, ModuleDataProvider, FileProvider, ModelUtils
from .state.config import Config
from .state.state import StateStore
from .types import PostRunHardwareState, DeckConfigurationType

from .engine_support import create_run_orchestrator


# TODO(mm, 2023-06-16): Arguably, this not being a context manager makes us prone to forgetting to
# clean it up properly, especially in tests. See e.g. https://opentrons.atlassian.net/browse/RSS-222
async def create_protocol_engine(
    hardware_api: HardwareControlAPI,
    config: Config,
    error_recovery_policy: ErrorRecoveryPolicy,
    load_fixed_trash: bool = False,
    deck_configuration: typing.Optional[DeckConfigurationType] = None,
    file_provider: typing.Optional[FileProvider] = None,
    notify_publishers: typing.Optional[typing.Callable[[], None]] = None,
) -> ProtocolEngine:
    """Create a ProtocolEngine instance.

    Arguments:
        hardware_api: Hardware control API to pass down to dependencies.
        config: ProtocolEngine configuration.
        error_recovery_policy: The error recovery policy to create the engine with.
            See documentation on `ErrorRecoveryPolicy`.
        load_fixed_trash: Automatically load fixed trash labware in engine.
        deck_configuration: The initial deck configuration the engine will be instantiated with.
        file_provider: Provides access to robot server file writing procedures for protocol output.
        notify_publishers: Notifies robot server publishers of internal state change.
    """
    deck_data = DeckDataProvider(config.deck_type)
    deck_definition = await deck_data.get_deck_definition()
    deck_fixed_labware = await deck_data.get_deck_fixed_labware(
        load_fixed_trash, deck_definition, deck_configuration
    )

    module_calibration_offsets = ModuleDataProvider.load_module_calibrations()
    robot_definition = load_robot(config.robot_type)

    state_store = StateStore(
        config=config,
        deck_definition=deck_definition,
        deck_fixed_labware=deck_fixed_labware,
        robot_definition=robot_definition,
        is_door_open=hardware_api.door_state is DoorState.OPEN,
        error_recovery_policy=error_recovery_policy,
        module_calibration_offsets=module_calibration_offsets,
        deck_configuration=deck_configuration,
        notify_publishers=notify_publishers,
    )
    hardware_state_synchronizer = ErrorRecoveryHardwareStateSynchronizer(
        hardware_api, state_store
    )
    action_dispatcher = ActionDispatcher(state_store)
    action_dispatcher.add_handler(hardware_state_synchronizer)
    plugin_starter = PluginStarter(state_store, action_dispatcher)
    model_utils = ModelUtils()
    hardware_stopper = HardwareStopper(hardware_api, state_store)
    door_watcher = DoorWatcher(state_store, hardware_api, action_dispatcher)
    module_data_provider = ModuleDataProvider()
    file_provider = file_provider or FileProvider()

    pe = ProtocolEngine(
        hardware_api=hardware_api,
        state_store=state_store,
        action_dispatcher=action_dispatcher,
        plugin_starter=plugin_starter,
        model_utils=model_utils,
        hardware_stopper=hardware_stopper,
        door_watcher=door_watcher,
        module_data_provider=module_data_provider,
        file_provider=file_provider,
    )

    # todo(mm, 2024-11-08): This is a quick hack to support the absorbance reader, which
    # expects the engine to have this special labware definition available. It would be
    # cleaner for the `loadModule` command to do this I/O and insert the definition
    # into state. That gets easier after https://opentrons.atlassian.net/browse/EXEC-756.
    #
    # NOTE: This needs to stay in sync with LabwareView.get_absorbance_reader_lid_definition().
    pe.add_labware_definition(
        await LabwareDataProvider().get_labware_definition(
            "opentrons_flex_lid_absorbance_plate_reader_module", "opentrons", 1
        )
    )

    return pe


@contextlib.contextmanager
def create_protocol_engine_in_thread(
    hardware_api: HardwareControlAPI,
    config: Config,
    deck_configuration: typing.Optional[DeckConfigurationType],
    file_provider: typing.Optional[FileProvider],
    error_recovery_policy: ErrorRecoveryPolicy,
    drop_tips_after_run: bool,
    post_run_hardware_state: PostRunHardwareState,
    load_fixed_trash: bool = False,
) -> typing.Generator[
    typing.Tuple[ProtocolEngine, asyncio.AbstractEventLoop], None, None
]:
    """Run a `ProtocolEngine` in a worker thread.

    When this context manager is entered, it:

    1. Starts a worker thread.
    2. Starts an asyncio event loop in that worker thread.
    3. Creates and `.play()`s a `ProtocolEngine` in that event loop.
    4. Returns the `ProtocolEngine` and the event loop.
       Use functions like `asyncio.run_coroutine_threadsafe()` to safely interact with
       the `ProtocolEngine` from your thread.

    When this context manager is exited, it:

    1. Cleans up the `ProtocolEngine`.
    2. Stops and cleans up the event loop.
    3. Joins the thread.
    """
    with async_context_manager_in_thread(
        _protocol_engine(
            hardware_api,
            file_provider,
            config,
            deck_configuration,
            error_recovery_policy,
            drop_tips_after_run,
            post_run_hardware_state,
            load_fixed_trash,
        )
    ) as (
        protocol_engine,
        loop,
    ):
        yield protocol_engine, loop


@contextlib.asynccontextmanager
async def _protocol_engine(
    hardware_api: HardwareControlAPI,
    file_provider: typing.Optional[FileProvider],
    config: Config,
    deck_configuration: typing.Optional[DeckConfigurationType],
    error_recovery_policy: ErrorRecoveryPolicy,
    drop_tips_after_run: bool,
    post_run_hardware_state: PostRunHardwareState,
    load_fixed_trash: bool = False,
) -> typing.AsyncGenerator[ProtocolEngine, None]:
    protocol_engine = await create_protocol_engine(
        hardware_api=hardware_api,
        file_provider=file_provider,
        config=config,
        error_recovery_policy=error_recovery_policy,
        load_fixed_trash=load_fixed_trash,
    )

    # TODO(tz, 6-20-2024): This feels like a hack, we should probably return the orchestrator instead of pe.
    orchestrator = create_run_orchestrator(
        hardware_api=hardware_api,
        protocol_engine=protocol_engine,
    )
    try:
        orchestrator.play(deck_configuration)
        yield protocol_engine
    finally:
        await orchestrator.finish(
            drop_tips_after_run=drop_tips_after_run,
            post_run_hardware_state=post_run_hardware_state,
        )



from asyncio import AbstractEventLoop
from typing_extensions import Protocol


class AsyncioConfigurable(Protocol):
    """Protocol specifying controllability of asyncio behavior"""

    @property
    def loop(self) -> AbstractEventLoop:
        """The event loop used by this instance."""
        ...



from dataclasses import dataclass
from typing import Optional, Sequence

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
    AspirateProperties,
    SingleDispenseProperties,
    MultiDispenseProperties,
    ByPipetteSetting,
    ByTipTypeSetting,
)


@dataclass(frozen=True)
class Liquid:
    """A liquid to load into a well.

    Attributes:
        name: A human-readable name for the liquid.
        description: An optional description.
        display_color: An optional display color for the liquid.

    .. versionadded:: 2.14
    """

    _id: str
    name: str
    description: Optional[str]
    display_color: Optional[str]


# TODO (spp, 2024-10-17): create PAPI-equivalent types for all the properties
#  and have validation on value updates with user-facing error messages
@dataclass
class TransferProperties:
    _aspirate: AspirateProperties
    _dispense: SingleDispenseProperties
    _multi_dispense: Optional[MultiDispenseProperties]

    @property
    def aspirate(self) -> AspirateProperties:
        """Aspirate properties."""
        return self._aspirate

    @property
    def dispense(self) -> SingleDispenseProperties:
        """Single dispense properties."""
        return self._dispense

    @property
    def multi_dispense(self) -> Optional[MultiDispenseProperties]:
        """Multi dispense properties."""
        return self._multi_dispense


@dataclass
class LiquidClass:
    """A data class that contains properties of a specific class of liquids."""

    _name: str
    _display_name: str
    _by_pipette_setting: Sequence[ByPipetteSetting]

    @classmethod
    def create(cls, liquid_class_definition: LiquidClassSchemaV1) -> "LiquidClass":
        """Liquid class factory method."""

        return cls(
            _name=liquid_class_definition.liquidClassName,
            _display_name=liquid_class_definition.displayName,
            _by_pipette_setting=liquid_class_definition.byPipette,
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def display_name(self) -> str:
        return self._display_name

    def get_for(self, pipette: str, tiprack: str) -> TransferProperties:
        """Get liquid class transfer properties for the specified pipette and tip."""
        settings_for_pipette: Sequence[ByPipetteSetting] = [
            pip_setting
            for pip_setting in self._by_pipette_setting
            if pip_setting.pipetteModel == pipette
        ]
        if len(settings_for_pipette) == 0:
            raise ValueError(
                f"No properties found for {pipette} in {self._name} liquid class"
            )
        settings_for_tip: Sequence[ByTipTypeSetting] = [
            tip_setting
            for tip_setting in settings_for_pipette[0].byTipType
            if tip_setting.tiprack == tiprack
        ]
        if len(settings_for_tip) == 0:
            raise ValueError(
                f"No properties found for {tiprack} in {self._name} liquid class"
            )
        return TransferProperties(
            _aspirate=settings_for_tip[0].aspirate,
            _dispense=settings_for_tip[0].singleDispense,
            _multi_dispense=settings_for_tip[0].multiDispense,
        )



from typing import Protocol, Sequence

from ..types import Axis


class PositionEstimator(Protocol):
    """Position-control extensions for harwdare with encoders."""

    async def update_axis_position_estimations(self, axes: Sequence[Axis]) -> None:
        """Update the specified axes' position estimators from their encoders.

        This will allow these axes to make a non-home move even if they do not currently have
        a position estimation (unless there is no tracked position from the encoders, as would be
        true immediately after boot).

        Axis encoders have less precision than their position estimators. Calling this function will
        cause absolute position drift. After this function is called, the axis should be homed before
        it is relied upon for accurate motion.

        This function updates only the requested axes. If other axes have bad position estimation,
        moves that require those axes or attempts to get the position of those axes will still fail.
        Axes that are not currently available (like a plunger for a pipette that is not connected)
        will be ignored.
        """
        ...

    def motor_status_ok(self, axis: Axis) -> bool:
        """Return whether an axis' position estimator is healthy.

        The position estimator is healthy if the axis has
        1) been homed
        2) not suffered a loss-of-positioning (from a cancel or stall, for instance) since being homed

        If this function returns false, getting the position of this axis or asking it to move will fail.
        """
        ...

    def encoder_status_ok(self, axis: Axis) -> bool:
        """Return whether an axis' position encoder tracking is healthy.

        The encoder status is healthy if the axis has been homed since booting up.

        If this function returns false, updating the estimator from the encoder will fail.
        """
        ...



"""Basic well data state and store."""
from dataclasses import dataclass
from typing import Dict, List, Union, Iterator, Optional, Tuple, overload, TypeVar

from opentrons.protocol_engine.types import (
    ProbedHeightInfo,
    ProbedVolumeInfo,
    LoadedVolumeInfo,
    WellInfoSummary,
    WellLiquidInfo,
)

from . import update_types
from ._abstract_store import HasState, HandlesActions
from ..actions import Action
from ..actions.get_state_update import get_state_updates


LabwareId = str
WellName = str


@dataclass
class WellState:
    """State of all wells."""

    loaded_volumes: Dict[LabwareId, Dict[WellName, LoadedVolumeInfo]]
    probed_heights: Dict[LabwareId, Dict[WellName, ProbedHeightInfo]]
    probed_volumes: Dict[LabwareId, Dict[WellName, ProbedVolumeInfo]]


class WellStore(HasState[WellState], HandlesActions):
    """Well state container."""

    _state: WellState

    def __init__(self) -> None:
        """Initialize a well store and its state."""
        self._state = WellState(loaded_volumes={}, probed_heights={}, probed_volumes={})

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            if state_update.liquid_loaded != update_types.NO_CHANGE:
                self._handle_liquid_loaded_update(state_update.liquid_loaded)
            if state_update.liquid_probed != update_types.NO_CHANGE:
                self._handle_liquid_probed_update(state_update.liquid_probed)
            if state_update.liquid_operated != update_types.NO_CHANGE:
                self._handle_liquid_operated_update(state_update.liquid_operated)

    def _handle_liquid_loaded_update(
        self, state_update: update_types.LiquidLoadedUpdate
    ) -> None:
        labware_id = state_update.labware_id
        if labware_id not in self._state.loaded_volumes:
            self._state.loaded_volumes[labware_id] = {}
        for (well, volume) in state_update.volumes.items():
            self._state.loaded_volumes[labware_id][well] = LoadedVolumeInfo(
                volume=_none_from_clear(volume),
                last_loaded=state_update.last_loaded,
                operations_since_load=0,
            )

    def _handle_liquid_probed_update(
        self, state_update: update_types.LiquidProbedUpdate
    ) -> None:
        labware_id = state_update.labware_id
        well_name = state_update.well_name
        if labware_id not in self._state.probed_heights:
            self._state.probed_heights[labware_id] = {}
        if labware_id not in self._state.probed_volumes:
            self._state.probed_volumes[labware_id] = {}
        self._state.probed_heights[labware_id][well_name] = ProbedHeightInfo(
            height=_none_from_clear(state_update.height),
            last_probed=state_update.last_probed,
        )
        self._state.probed_volumes[labware_id][well_name] = ProbedVolumeInfo(
            volume=_none_from_clear(state_update.volume),
            last_probed=state_update.last_probed,
            operations_since_probe=0,
        )

    def _handle_liquid_operated_update(
        self, state_update: update_types.LiquidOperatedUpdate
    ) -> None:
        labware_id = state_update.labware_id
        well_name = state_update.well_name
        if (
            labware_id in self._state.loaded_volumes
            and well_name in self._state.loaded_volumes[labware_id]
        ):
            if state_update.volume_added is update_types.CLEAR:
                del self._state.loaded_volumes[labware_id][well_name]
            else:
                prev_loaded_vol_info = self._state.loaded_volumes[labware_id][well_name]
                assert prev_loaded_vol_info.volume is not None
                self._state.loaded_volumes[labware_id][well_name] = LoadedVolumeInfo(
                    volume=prev_loaded_vol_info.volume + state_update.volume_added,
                    last_loaded=prev_loaded_vol_info.last_loaded,
                    operations_since_load=prev_loaded_vol_info.operations_since_load
                    + 1,
                )
        if (
            labware_id in self._state.probed_heights
            and well_name in self._state.probed_heights[labware_id]
        ):
            del self._state.probed_heights[labware_id][well_name]
        if (
            labware_id in self._state.probed_volumes
            and well_name in self._state.probed_volumes[labware_id]
        ):
            if state_update.volume_added is update_types.CLEAR:
                del self._state.probed_volumes[labware_id][well_name]
            else:
                prev_probed_vol_info = self._state.probed_volumes[labware_id][well_name]
                if prev_probed_vol_info.volume is None:
                    new_vol_info: float | None = None
                else:
                    new_vol_info = (
                        prev_probed_vol_info.volume + state_update.volume_added
                    )
                self._state.probed_volumes[labware_id][well_name] = ProbedVolumeInfo(
                    volume=new_vol_info,
                    last_probed=prev_probed_vol_info.last_probed,
                    operations_since_probe=prev_probed_vol_info.operations_since_probe
                    + 1,
                )


class WellView(HasState[WellState]):
    """Read-only well state view."""

    _state: WellState

    def __init__(self, state: WellState) -> None:
        """Initialize the computed view of well state.

        Arguments:
            state: Well state dataclass used for all calculations.
        """
        self._state = state

    def get_well_liquid_info(self, labware_id: str, well_name: str) -> WellLiquidInfo:
        """Return all the liquid info for a well."""
        if (
            labware_id not in self._state.loaded_volumes
            or well_name not in self._state.loaded_volumes[labware_id]
        ):
            loaded_volume_info = None
        else:
            loaded_volume_info = self._state.loaded_volumes[labware_id][well_name]
        if (
            labware_id not in self._state.probed_heights
            or well_name not in self._state.probed_heights[labware_id]
        ):
            probed_height_info = None
        else:
            probed_height_info = self._state.probed_heights[labware_id][well_name]
        if (
            labware_id not in self._state.probed_volumes
            or well_name not in self._state.probed_volumes[labware_id]
        ):
            probed_volume_info = None
        else:
            probed_volume_info = self._state.probed_volumes[labware_id][well_name]
        return WellLiquidInfo(
            loaded_volume=loaded_volume_info,
            probed_height=probed_height_info,
            probed_volume=probed_volume_info,
        )

    def get_all(self) -> List[WellInfoSummary]:
        """Get all well liquid info summaries."""

        def _all_well_combos() -> Iterator[Tuple[str, str, str]]:
            for labware, lv_wells in self._state.loaded_volumes.items():
                for well_name in lv_wells.keys():
                    yield f"{labware}{well_name}", labware, well_name
            for labware, ph_wells in self._state.probed_heights.items():
                for well_name in ph_wells.keys():
                    yield f"{labware}{well_name}", labware, well_name
            for labware, pv_wells in self._state.probed_volumes.items():
                for well_name in pv_wells.keys():
                    yield f"{labware}{well_name}", labware, well_name

        wells = {
            key: (labware_id, well_name)
            for key, labware_id, well_name in _all_well_combos()
        }
        return [
            self._summarize_well(labware_id, well_name)
            for labware_id, well_name in wells.values()
        ]

    def _summarize_well(self, labware_id: str, well_name: str) -> WellInfoSummary:
        well_liquid_info = self.get_well_liquid_info(labware_id, well_name)
        return WellInfoSummary(
            labware_id=labware_id,
            well_name=well_name,
            loaded_volume=_volume_from_info(well_liquid_info.loaded_volume),
            probed_volume=_volume_from_info(well_liquid_info.probed_volume),
            probed_height=_height_from_info(well_liquid_info.probed_height),
        )


@overload
def _volume_from_info(info: Optional[ProbedVolumeInfo]) -> Optional[float]:
    ...


@overload
def _volume_from_info(info: Optional[LoadedVolumeInfo]) -> Optional[float]:
    ...


def _volume_from_info(
    info: Union[ProbedVolumeInfo, LoadedVolumeInfo, None]
) -> Optional[float]:
    if info is None:
        return None
    return info.volume


def _height_from_info(info: Optional[ProbedHeightInfo]) -> Optional[float]:
    if info is None:
        return None
    return info.height


MaybeClear = TypeVar("MaybeClear")


def _none_from_clear(inval: MaybeClear | update_types.ClearType) -> MaybeClear | None:
    if inval == update_types.CLEAR:
        return None
    return inval



"""Custom protocol command.

A "custom" command contains arbitrary payload and result data.
It can be used by ProtocolEngine plugins to represent external
commands in state that a vanilla ProtocolEngine would not
know about.

This data model serves as a wrapper to ensure custom, arbitrary
data still adheres to the shapes that ProtocolEngine expects.
If you are implementing a custom command, you should probably
put your own disambiguation identifier in the payload.
"""
from pydantic import BaseModel, Extra
from typing import Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence


CustomCommandType = Literal["custom"]


class CustomParams(BaseModel):
    """Payload used by a custom command."""

    class Config:
        """Allow arbitrary fields."""

        extra = Extra.allow


class CustomResult(BaseModel):
    """Result data from a custom command."""

    class Config:
        """Allow arbitrary fields."""

        extra = Extra.allow


class CustomImplementation(
    AbstractCommandImpl[CustomParams, SuccessData[CustomResult]]
):
    """Custom command implementation."""

    # TODO(mm, 2022-11-09): figure out how a plugin can specify a custom command
    # implementation. For now, always no-op, so we can use custom commands as containers
    # for legacy RPC (pre-ProtocolEngine) payloads.
    async def execute(self, params: CustomParams) -> SuccessData[CustomResult]:
        """A custom command does nothing when executed directly."""
        return SuccessData(
            public=CustomResult.construct(),
        )


class Custom(BaseCommand[CustomParams, CustomResult, ErrorOccurrence]):
    """Custom command model."""

    commandType: CustomCommandType = "custom"
    params: CustomParams
    result: Optional[CustomResult]

    _ImplementationCls: Type[CustomImplementation] = CustomImplementation


class CustomCreate(BaseCommandCreate[CustomParams]):
    """A request to create a custom command."""

    commandType: CustomCommandType = "custom"
    params: CustomParams

    _CommandCls: Type[Custom] = Custom
